# test_commit_graph.py -- Compatibility tests for commit graph functionality
# Copyright (C) 2024 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.

"""Compatibility tests for Git commit graph functionality.

These tests verify that dulwich's commit graph implementation behaves
identically to C Git's implementation.
"""

import os
import tempfile

from dulwich.commit_graph import find_commit_graph_file, read_commit_graph
from dulwich.graph import can_fast_forward, find_merge_base
from dulwich.repo import Repo

from .utils import CompatTestCase, run_git_or_fail


class CommitGraphCompatTests(CompatTestCase):
    """Compatibility tests for commit graph functionality."""

    # Commit graph was introduced in Git 2.18.0
    min_git_version = (2, 18, 0)

    def setUp(self):
        super().setUp()
        self.test_dir = tempfile.mkdtemp()
        self.repo_path = os.path.join(self.test_dir, "test-repo")

        # Set up git identity to avoid committer identity errors
        self.overrideEnv("GIT_COMMITTER_NAME", "Test Author")
        self.overrideEnv("GIT_COMMITTER_EMAIL", "test@example.com")
        self.overrideEnv("GIT_AUTHOR_NAME", "Test Author")
        self.overrideEnv("GIT_AUTHOR_EMAIL", "test@example.com")

    def tearDown(self):
        from .utils import rmtree_ro

        rmtree_ro(self.test_dir)

    def create_test_repo_with_history(self):
        """Create a test repository with some commit history."""
        # Initialize repository
        run_git_or_fail(["init"], cwd=self.test_dir)
        os.rename(os.path.join(self.test_dir, ".git"), self.repo_path)

        work_dir = os.path.join(self.test_dir, "work")
        os.makedirs(work_dir)

        # Create .git file pointing to our repo
        with open(os.path.join(work_dir, ".git"), "w") as f:
            f.write(f"gitdir: {self.repo_path}\n")

        # Create some commits
        commits = []
        for i in range(5):
            filename = f"file{i}.txt"
            with open(os.path.join(work_dir, filename), "w") as f:
                f.write(f"Content {i}\n")

            run_git_or_fail(["add", filename], cwd=work_dir)
            run_git_or_fail(
                [
                    "commit",
                    "-m",
                    f"Commit {i}",
                    "--author",
                    "Test Author <test@example.com>",
                    "--date",
                    f"2024-01-0{i + 1} 12:00:00 +0000",
                ],
                cwd=work_dir,
            )

            # Get the commit SHA
            result = run_git_or_fail(["rev-parse", "HEAD"], cwd=work_dir)
            commits.append(result.strip())

        # Create a branch and merge
        run_git_or_fail(["checkout", "-b", "feature"], cwd=work_dir)

        with open(os.path.join(work_dir, "feature.txt"), "w") as f:
            f.write("Feature content\n")
        run_git_or_fail(["add", "feature.txt"], cwd=work_dir)
        run_git_or_fail(
            [
                "commit",
                "-m",
                "Feature commit",
                "--author",
                "Test Author <test@example.com>",
                "--date",
                "2024-01-06 12:00:00 +0000",
            ],
            cwd=work_dir,
        )

        result = run_git_or_fail(["rev-parse", "HEAD"], cwd=work_dir)
        feature_commit = result.strip()

        # Merge back to master
        run_git_or_fail(["checkout", "master"], cwd=work_dir)
        run_git_or_fail(
            ["merge", "feature", "--no-ff", "-m", "Merge feature"], cwd=work_dir
        )

        result = run_git_or_fail(["rev-parse", "HEAD"], cwd=work_dir)
        merge_commit = result.strip()

        commits.extend([feature_commit, merge_commit])
        return commits, work_dir

    def test_commit_graph_generation_and_reading(self):
        """Test that dulwich can read commit graphs generated by C Git."""
        _commits, work_dir = self.create_test_repo_with_history()

        # Generate commit graph with C Git
        run_git_or_fail(["commit-graph", "write", "--reachable"], cwd=work_dir)

        # Verify commit graph file exists
        graph_file = find_commit_graph_file(self.repo_path)
        self.assertIsNotNone(graph_file, "Commit graph file should exist")

        # Read with dulwich
        commit_graph = read_commit_graph(graph_file)
        self.assertIsNotNone(commit_graph, "Should be able to read commit graph")

        # Verify we have the expected number of commits
        self.assertGreater(len(commit_graph), 0, "Commit graph should contain commits")

        # Open the repository with dulwich
        repo = Repo(self.repo_path)
        self.addCleanup(repo.close)

        # Verify that all commits in the graph are accessible
        for entry in commit_graph:
            # entry.commit_id is hex ObjectID
            commit_obj = repo.object_store[entry.commit_id]
            self.assertIsNotNone(
                commit_obj, f"Commit {entry.commit_id.decode()} should be accessible"
            )

            # Verify tree ID matches
            self.assertEqual(
                entry.tree_id,
                commit_obj.tree,
                f"Tree ID mismatch for commit {entry.commit_id.decode()}",
            )

            # Verify parent information
            self.assertEqual(
                entry.parents,
                list(commit_obj.parents),
                f"Parent mismatch for commit {entry.commit_id.decode()}",
            )

    def test_merge_base_with_commit_graph(self):
        """Test that merge-base calculations work the same with and without commit graph."""
        _commits, work_dir = self.create_test_repo_with_history()

        repo = Repo(self.repo_path)
        self.addCleanup(repo.close)

        # Get some commit IDs for testing
        main_head = repo.refs[b"refs/heads/master"]
        feature_head = repo.refs[b"refs/heads/feature"]

        # Calculate merge base without commit graph
        merge_base_no_graph = find_merge_base(repo, [main_head, feature_head])

        # Generate commit graph
        run_git_or_fail(["commit-graph", "write", "--reachable"], cwd=work_dir)

        # Force reload of repository to pick up commit graph
        repo.close()
        repo = Repo(self.repo_path)
        self.addCleanup(repo.close)

        # Calculate merge base with commit graph
        merge_base_with_graph = find_merge_base(repo, [main_head, feature_head])

        # Results should be identical
        self.assertEqual(
            merge_base_no_graph,
            merge_base_with_graph,
            "Merge base should be same with and without commit graph",
        )

        # Compare with C Git's result
        git_result = run_git_or_fail(
            ["merge-base", main_head.decode(), feature_head.decode()], cwd=work_dir
        )
        git_merge_base = [git_result.strip()]

        self.assertEqual(
            merge_base_with_graph,
            git_merge_base,
            "Dulwich merge base should match C Git result",
        )

    def test_fast_forward_with_commit_graph(self):
        """Test that fast-forward detection works the same with and without commit graph."""
        commits, work_dir = self.create_test_repo_with_history()

        repo = Repo(self.repo_path)
        self.addCleanup(repo.close)

        # Test with a simple fast-forward case (older commit to newer commit)
        commit1 = commits[1]  # Second commit
        commit2 = commits[3]  # Fourth commit

        # Check without commit graph
        can_ff_no_graph = can_fast_forward(repo, commit1, commit2)

        # Generate commit graph
        run_git_or_fail(["commit-graph", "write", "--reachable"], cwd=work_dir)

        # Force reload
        repo.close()
        repo = Repo(self.repo_path)
        self.addCleanup(repo.close)

        # Check with commit graph
        can_ff_with_graph = can_fast_forward(repo, commit1, commit2)

        # Results should be identical
        self.assertEqual(
            can_ff_no_graph,
            can_ff_with_graph,
            "Fast-forward detection should be same with and without commit graph",
        )

        # Compare with C Git (check if commit1 is ancestor of commit2)
        from .utils import run_git

        returncode, _stdout, _stderr = run_git(
            ["merge-base", "--is-ancestor", commit1.decode(), commit2.decode()],
            cwd=work_dir,
            capture_stdout=True,
            capture_stderr=True,
        )
        git_can_ff = returncode == 0

        self.assertEqual(
            can_ff_with_graph,
            git_can_ff,
            "Dulwich fast-forward detection should match C Git",
        )

    def test_generation_numbers_consistency(self):
        """Test that generation numbers are consistent with Git's topology."""
        _commits, work_dir = self.create_test_repo_with_history()

        # Generate commit graph
        run_git_or_fail(["commit-graph", "write", "--reachable"], cwd=work_dir)

        # Read with dulwich
        graph_file = find_commit_graph_file(self.repo_path)
        commit_graph = read_commit_graph(graph_file)

        repo = Repo(self.repo_path)
        self.addCleanup(repo.close)

        # Build a map of commit to generation number
        generation_map = {}
        for entry in commit_graph:
            generation_map[entry.commit_id] = entry.generation

        # Verify generation number properties:
        # 1. Root commits have generation 1
        # 2. For any commit, generation > max(parent_generations)
        for entry in commit_graph:
            repo.object_store[entry.commit_id]

            if not entry.parents:
                # Root commit should have generation 1
                self.assertGreaterEqual(
                    entry.generation,
                    1,
                    f"Root commit {entry.commit_id.decode()} should have generation >= 1",
                )
            else:
                # Non-root commit should have generation > max parent generation
                max_parent_gen = 0
                for parent_id in entry.parents:
                    if parent_id in generation_map:
                        max_parent_gen = max(max_parent_gen, generation_map[parent_id])

                if max_parent_gen > 0:  # Only check if we have parent generation info
                    self.assertGreater(
                        entry.generation,
                        max_parent_gen,
                        f"Commit {entry.commit_id.decode()} generation should be > max parent generation",
                    )

    def test_commit_graph_with_different_options(self):
        """Test commit graph generation with different C Git options."""
        commits, work_dir = self.create_test_repo_with_history()

        # Test different generation strategies
        strategies = [
            ["--reachable"],
            ["--stdin-commits"],
            ["--append"],
        ]

        for i, strategy in enumerate(strategies):
            with self.subTest(strategy=strategy):
                # Clean up any existing commit graph
                graph_path = os.path.join(
                    self.repo_path, "objects", "info", "commit-graph"
                )
                if os.path.exists(graph_path):
                    try:
                        os.remove(graph_path)
                    except PermissionError:
                        # On Windows, handle read-only files
                        from .utils import remove_ro

                        remove_ro(graph_path)

                if strategy == ["--stdin-commits"]:
                    # For stdin-commits, we need to provide commit IDs
                    process_input = b"\n".join(commits[:3]) + b"\n"  # First 3 commits
                    run_git_or_fail(
                        ["commit-graph", "write", *strategy],
                        cwd=work_dir,
                        input=process_input,
                    )
                elif strategy == ["--append"]:
                    # First create a base graph
                    run_git_or_fail(
                        ["commit-graph", "write", "--reachable"], cwd=work_dir
                    )
                    # Then append (this should work even if nothing new to add)
                    run_git_or_fail(["commit-graph", "write", *strategy], cwd=work_dir)
                else:
                    run_git_or_fail(["commit-graph", "write", *strategy], cwd=work_dir)

                # Verify dulwich can read the generated graph
                graph_file = find_commit_graph_file(self.repo_path)
                if graph_file:  # Some strategies might not generate a graph
                    commit_graph = read_commit_graph(graph_file)
                    self.assertIsNotNone(
                        commit_graph,
                        f"Should be able to read commit graph generated with {strategy}",
                    )

    def test_commit_graph_incremental_update(self):
        """Test that dulwich can read incrementally updated commit graphs."""
        _commits, work_dir = self.create_test_repo_with_history()

        # Create initial commit graph
        run_git_or_fail(["commit-graph", "write", "--reachable"], cwd=work_dir)

        # Read initial graph
        graph_file = find_commit_graph_file(self.repo_path)
        initial_graph = read_commit_graph(graph_file)
        initial_count = len(initial_graph)

        # Add another commit
        with open(os.path.join(work_dir, "new_file.txt"), "w") as f:
            f.write("New content\n")
        run_git_or_fail(["add", "new_file.txt"], cwd=work_dir)
        run_git_or_fail(
            [
                "commit",
                "-m",
                "New commit",
                "--author",
                "Test Author <test@example.com>",
                "--date",
                "2024-01-08 12:00:00 +0000",
            ],
            cwd=work_dir,
        )

        # Update commit graph
        run_git_or_fail(["commit-graph", "write", "--reachable"], cwd=work_dir)

        # Read updated graph
        updated_graph = read_commit_graph(graph_file)

        # Should have one more commit
        self.assertEqual(
            len(updated_graph),
            initial_count + 1,
            "Updated commit graph should have one more commit",
        )

        # Verify all original commits are still present
        initial_commit_ids = {entry.commit_id for entry in initial_graph}
        updated_commit_ids = {entry.commit_id for entry in updated_graph}

        self.assertTrue(
            initial_commit_ids.issubset(updated_commit_ids),
            "All original commits should still be in updated graph",
        )

    def test_commit_graph_with_tags_and_refs(self):
        """Test commit graph behavior with various refs and tags."""
        commits, work_dir = self.create_test_repo_with_history()

        # Create some tags
        run_git_or_fail(["tag", "v1.0", commits[2].decode()], cwd=work_dir)
        run_git_or_fail(
            ["tag", "-a", "v2.0", commits[4].decode(), "-m", "Version 2.0"],
            cwd=work_dir,
        )

        # Generate commit graph
        run_git_or_fail(["commit-graph", "write", "--reachable"], cwd=work_dir)

        # Verify dulwich can read the graph
        graph_file = find_commit_graph_file(self.repo_path)
        commit_graph = read_commit_graph(graph_file)

        repo = Repo(self.repo_path)
        self.addCleanup(repo.close)

        # Verify tagged commits are in the graph
        tagged_commits = [commits[2], commits[4]]
        graph_commit_ids = {entry.commit_id for entry in commit_graph}

        for tagged_commit in tagged_commits:
            self.assertIn(
                tagged_commit,
                graph_commit_ids,
                f"Tagged commit {tagged_commit.decode()} should be in commit graph",
            )

        # Test merge base with tagged commits
        merge_base = find_merge_base(repo, [commits[0], commits[4]])
        self.assertEqual(
            merge_base,
            [commits[0]],
            "Merge base calculation should work with tagged commits",
        )

    def test_empty_repository_commit_graph(self):
        """Test commit graph behavior with empty repository."""
        # Create empty repository
        run_git_or_fail(["init", "--bare"], cwd=self.test_dir)
        empty_repo_path = os.path.join(self.test_dir, ".git")

        # Try to write commit graph (should succeed but create empty graph)
        try:
            run_git_or_fail(["commit-graph", "write", "--reachable"], cwd=self.test_dir)
        except AssertionError:
            # Some Git versions might fail on empty repos, which is fine
            pass

        # Check if commit graph file exists
        graph_file = find_commit_graph_file(empty_repo_path)
        if graph_file:
            # If it exists, dulwich should be able to read it
            commit_graph = read_commit_graph(graph_file)
            self.assertIsNotNone(
                commit_graph, "Should be able to read empty commit graph"
            )
            self.assertEqual(
                len(commit_graph), 0, "Empty repository should have empty commit graph"
            )



# test_bundle_uri.py -- tests for bundle URI support
# Copyright (C) 2024 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for bundle URI support."""

from dulwich.bundle_uri import (
    BundleList,
    BundleListEntry,
    BundleURIError,
    _filter_bundle_entries,
    _is_absolute_uri,
    _resolve_relative_uri,
    parse_bundle_list,
)

from . import TestCase


class BundleURIParsingTests(TestCase):
    def test_parse_simple_bundle_list(self) -> None:
        """Test parsing a simple bundle list."""
        data = b"""[bundle]
    version = 1
    mode = all

[bundle "bundle1"]
    uri = https://example.com/bundle1.bundle

[bundle "bundle2"]
    uri = https://example.com/bundle2.bundle
"""
        bundle_list = parse_bundle_list(data)

        self.assertEqual(bundle_list.version, 1)
        self.assertEqual(bundle_list.mode, "all")
        self.assertIsNone(bundle_list.heuristic)
        self.assertEqual(len(bundle_list.entries), 2)

        # Check entries
        entry_ids = {e.id for e in bundle_list.entries}
        self.assertIn("bundle1", entry_ids)
        self.assertIn("bundle2", entry_ids)

    def test_parse_bundle_list_with_heuristic(self) -> None:
        """Test parsing a bundle list with creationToken heuristic."""
        data = b"""[bundle]
    version = 1
    mode = all
    heuristic = creationToken

[bundle "daily-2024"]
    uri = https://example.com/daily.bundle
    creationToken = 1700000000

[bundle "weekly-2024"]
    uri = https://example.com/weekly.bundle
    creationToken = 1699000000
"""
        bundle_list = parse_bundle_list(data)

        self.assertEqual(bundle_list.heuristic, "creationToken")
        self.assertEqual(len(bundle_list.entries), 2)

        # Find daily entry
        daily = next(e for e in bundle_list.entries if e.id == "daily-2024")
        self.assertEqual(daily.creation_token, 1700000000)

        weekly = next(e for e in bundle_list.entries if e.id == "weekly-2024")
        self.assertEqual(weekly.creation_token, 1699000000)

    def test_parse_bundle_list_any_mode(self) -> None:
        """Test parsing a bundle list with 'any' mode."""
        data = b"""[bundle]
    version = 1
    mode = any

[bundle "eastus"]
    uri = https://eastus.example.com/bundle.bundle
    location = eastus

[bundle "westus"]
    uri = https://westus.example.com/bundle.bundle
    location = westus
"""
        bundle_list = parse_bundle_list(data)

        self.assertEqual(bundle_list.mode, "any")
        self.assertEqual(len(bundle_list.entries), 2)

        eastus = next(e for e in bundle_list.entries if e.id == "eastus")
        self.assertEqual(eastus.location, "eastus")

    def test_parse_bundle_list_with_filter(self) -> None:
        """Test parsing a bundle list with object filters."""
        data = b"""[bundle]
    version = 1
    mode = all

[bundle "full"]
    uri = https://example.com/full.bundle

[bundle "blobless"]
    uri = https://example.com/blobless.bundle
    filter = blob:none
"""
        bundle_list = parse_bundle_list(data)

        full = next(e for e in bundle_list.entries if e.id == "full")
        self.assertIsNone(full.filter)

        blobless = next(e for e in bundle_list.entries if e.id == "blobless")
        self.assertEqual(blobless.filter, "blob:none")

    def test_parse_bundle_list_missing_version(self) -> None:
        """Test that missing version raises error."""
        data = b"""[bundle]
    mode = all
"""
        with self.assertRaises(BundleURIError) as cm:
            parse_bundle_list(data)
        self.assertIn("version", str(cm.exception))

    def test_parse_bundle_list_missing_mode(self) -> None:
        """Test that missing mode raises error."""
        data = b"""[bundle]
    version = 1
"""
        with self.assertRaises(BundleURIError) as cm:
            parse_bundle_list(data)
        self.assertIn("mode", str(cm.exception))

    def test_parse_bundle_list_invalid_mode(self) -> None:
        """Test that invalid mode raises error."""
        data = b"""[bundle]
    version = 1
    mode = invalid
"""
        with self.assertRaises(BundleURIError) as cm:
            parse_bundle_list(data)
        self.assertIn("mode", str(cm.exception))

    def test_parse_bundle_list_unsupported_version(self) -> None:
        """Test that unsupported version raises error."""
        data = b"""[bundle]
    version = 2
    mode = all
"""
        with self.assertRaises(BundleURIError) as cm:
            parse_bundle_list(data)
        self.assertIn("version", str(cm.exception).lower())

    def test_parse_bundle_list_relative_uri(self) -> None:
        """Test parsing bundle list with relative URIs."""
        data = b"""[bundle]
    version = 1
    mode = all

[bundle "relative"]
    uri = daily.bundle

[bundle "absolute-path"]
    uri = /bundles/weekly.bundle
"""
        base_uri = "https://example.com/git/repo/"
        bundle_list = parse_bundle_list(data, base_uri=base_uri)

        relative = next(e for e in bundle_list.entries if e.id == "relative")
        self.assertEqual(relative.uri, "https://example.com/git/repo/daily.bundle")

        absolute_path = next(e for e in bundle_list.entries if e.id == "absolute-path")
        self.assertEqual(absolute_path.uri, "https://example.com/bundles/weekly.bundle")


class URIResolutionTests(TestCase):
    def test_is_absolute_uri(self) -> None:
        """Test absolute URI detection."""
        self.assertTrue(_is_absolute_uri("https://example.com/bundle.bundle"))
        self.assertTrue(_is_absolute_uri("http://example.com/bundle.bundle"))
        self.assertFalse(_is_absolute_uri("/bundles/bundle.bundle"))
        self.assertFalse(_is_absolute_uri("daily.bundle"))
        self.assertFalse(_is_absolute_uri("../bundle.bundle"))

    def test_resolve_relative_uri_simple(self) -> None:
        """Test resolving simple relative URIs."""
        base = "https://example.com/git/repo/"
        self.assertEqual(
            _resolve_relative_uri(base, "daily.bundle"),
            "https://example.com/git/repo/daily.bundle",
        )

    def test_resolve_relative_uri_absolute_path(self) -> None:
        """Test resolving absolute path URIs."""
        base = "https://example.com/git/repo/"
        self.assertEqual(
            _resolve_relative_uri(base, "/bundles/daily.bundle"),
            "https://example.com/bundles/daily.bundle",
        )

    def test_resolve_relative_uri_parent(self) -> None:
        """Test resolving parent-relative URIs."""
        base = "https://example.com/git/repo/"
        self.assertEqual(
            _resolve_relative_uri(base, "../other/daily.bundle"),
            "https://example.com/git/other/daily.bundle",
        )


class BundleFilteringTests(TestCase):
    def test_filter_entries_full_clone(self) -> None:
        """Test filtering entries for full clone (no filter)."""
        entries = [
            BundleListEntry(id="full", uri="https://example.com/full.bundle"),
            BundleListEntry(
                id="blobless",
                uri="https://example.com/blobless.bundle",
                filter="blob:none",
            ),
        ]

        filtered = _filter_bundle_entries(entries, filter_spec=None)
        self.assertEqual(len(filtered), 1)
        self.assertEqual(filtered[0].id, "full")

    def test_filter_entries_partial_clone(self) -> None:
        """Test filtering entries for partial clone with filter."""
        entries = [
            BundleListEntry(id="full", uri="https://example.com/full.bundle"),
            BundleListEntry(
                id="blobless",
                uri="https://example.com/blobless.bundle",
                filter="blob:none",
            ),
            BundleListEntry(
                id="treeless",
                uri="https://example.com/treeless.bundle",
                filter="tree:0",
            ),
        ]

        filtered = _filter_bundle_entries(entries, filter_spec="blob:none")
        self.assertEqual(len(filtered), 1)
        self.assertEqual(filtered[0].id, "blobless")

    def test_filter_entries_no_match(self) -> None:
        """Test filtering when no entries match."""
        entries = [
            BundleListEntry(
                id="blobless",
                uri="https://example.com/blobless.bundle",
                filter="blob:none",
            ),
        ]

        filtered = _filter_bundle_entries(entries, filter_spec=None)
        self.assertEqual(len(filtered), 0)


class BundleListEntryTests(TestCase):
    def test_entry_defaults(self) -> None:
        """Test BundleListEntry default values."""
        entry = BundleListEntry(id="test", uri="https://example.com/test.bundle")
        self.assertEqual(entry.id, "test")
        self.assertEqual(entry.uri, "https://example.com/test.bundle")
        self.assertIsNone(entry.filter)
        self.assertIsNone(entry.creation_token)
        self.assertIsNone(entry.location)

    def test_entry_all_fields(self) -> None:
        """Test BundleListEntry with all fields set."""
        entry = BundleListEntry(
            id="test",
            uri="https://example.com/test.bundle",
            filter="blob:none",
            creation_token=1700000000,
            location="eastus",
        )
        self.assertEqual(entry.id, "test")
        self.assertEqual(entry.uri, "https://example.com/test.bundle")
        self.assertEqual(entry.filter, "blob:none")
        self.assertEqual(entry.creation_token, 1700000000)
        self.assertEqual(entry.location, "eastus")


class BundleListTests(TestCase):
    def test_bundle_list_defaults(self) -> None:
        """Test BundleList default values."""
        bundle_list = BundleList()
        self.assertEqual(bundle_list.version, 1)
        self.assertEqual(bundle_list.mode, "all")
        self.assertIsNone(bundle_list.heuristic)
        self.assertEqual(bundle_list.entries, [])

    def test_bundle_list_custom(self) -> None:
        """Test BundleList with custom values."""
        entries = [
            BundleListEntry(id="test", uri="https://example.com/test.bundle"),
        ]
        bundle_list = BundleList(
            version=1,
            mode="any",
            heuristic="creationToken",
            entries=entries,
        )
        self.assertEqual(bundle_list.version, 1)
        self.assertEqual(bundle_list.mode, "any")
        self.assertEqual(bundle_list.heuristic, "creationToken")
        self.assertEqual(len(bundle_list.entries), 1)



# test_aiohttp.py -- Compatibility tests for the aiohttp HTTP server.
# Copyright (C) 2025 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Compatibility tests between Dulwich and the cgit HTTP client using aiohttp server.

warning: these tests should be fairly stable, but when writing/debugging new
    tests, deadlocks may freeze the test process such that it cannot be
    Ctrl-C'ed. On POSIX systems, you can kill the tests with Ctrl-Z, "kill %".
"""

import asyncio
import sys
import threading
from typing import NoReturn

try:
    from aiohttp import web

    aiohttp_missing = False
except ImportError:
    web = None  # type: ignore
    aiohttp_missing = True

from dulwich.server import ReceivePackHandler

from .. import SkipTest, skipIf
from .server_utils import NoSideBand64kReceivePackHandler, ServerTests
from .utils import CompatTestCase

if not aiohttp_missing:
    from dulwich.aiohttp.server import create_repo_app


@skipIf(aiohttp_missing, "aiohttp not available")
@skipIf(sys.platform == "win32", "Broken on windows, with very long fail time.")
class AiohttpServerTests(ServerTests):
    """Base tests for aiohttp server tests.

    Contains utility and setUp/tearDown methods, but does not inherit from
    TestCase so tests are not automatically run.
    """

    protocol = "http"

    def _start_server(self, repo):
        app = self._make_app(repo)
        runner = web.AppRunner(app)

        loop = asyncio.new_event_loop()

        async def start():
            await runner.setup()
            site = web.TCPSite(runner, "localhost", 0)
            await site.start()
            return site

        site = loop.run_until_complete(start())

        # Get the actual port
        port = site._server.sockets[0].getsockname()[1]

        # Run the event loop in a separate thread
        def run_loop():
            asyncio.set_event_loop(loop)
            loop.run_forever()

        thread = threading.Thread(target=run_loop, daemon=True)
        thread.start()

        # Cleanup function
        def cleanup():
            async def stop():
                await site.stop()
                await runner.cleanup()

            future = asyncio.run_coroutine_threadsafe(stop(), loop)
            future.result(timeout=5)
            loop.call_soon_threadsafe(loop.stop)
            thread.join()
            loop.close()

        self.addCleanup(cleanup)
        self._server = runner
        return port


@skipIf(aiohttp_missing, "aiohttp not available")
@skipIf(sys.platform == "win32", "Broken on windows, with very long fail time.")
class SmartAiohttpTestCase(AiohttpServerTests, CompatTestCase):
    """Test cases for smart HTTP server using aiohttp.

    This server test case does not use side-band-64k in git-receive-pack.
    """

    min_git_version: tuple[int, ...] = (1, 6, 6)

    def _handlers(self):
        return {b"git-receive-pack": NoSideBand64kReceivePackHandler}

    def _make_app(self, repo):
        from dulwich.server import DEFAULT_HANDLERS

        handlers = dict(DEFAULT_HANDLERS)
        handlers.update(self._handlers())
        return create_repo_app(repo, handlers=handlers)


def patch_capabilities(handler, caps_removed):
    # Patch a handler's capabilities by specifying a list of them to be
    # removed, and return the original method for restoration.
    original_capabilities = handler.capabilities

    def capabilities(self):
        # Call original to get base capabilities (including object-format)
        base_caps = original_capabilities(self)
        # Filter out the capabilities we want to remove
        return [i for i in base_caps if i not in caps_removed]

    handler.capabilities = capabilities
    return original_capabilities


@skipIf(aiohttp_missing, "aiohttp not available")
@skipIf(sys.platform == "win32", "Broken on windows, with very long fail time.")
class SmartAiohttpSideBand64kTestCase(SmartAiohttpTestCase):
    """Test cases for smart HTTP server with side-band-64k support using aiohttp."""

    # side-band-64k in git-receive-pack was introduced in git 1.7.0.2
    min_git_version = (1, 7, 0, 2)

    def setUp(self) -> None:
        from dulwich.server import UploadPackHandler

        self.o_uph_cap = patch_capabilities(UploadPackHandler, (b"no-done",))
        self.o_rph_cap = patch_capabilities(ReceivePackHandler, (b"no-done",))
        super().setUp()

    def tearDown(self) -> None:
        from dulwich.server import UploadPackHandler

        super().tearDown()
        UploadPackHandler.capabilities = self.o_uph_cap
        ReceivePackHandler.capabilities = self.o_rph_cap

    def _make_app(self, repo):
        return create_repo_app(repo)


@skipIf(aiohttp_missing, "aiohttp not available")
@skipIf(sys.platform == "win32", "Broken on windows, with very long fail time.")
class SmartAiohttpSideBand64kNoDoneTestCase(SmartAiohttpTestCase):
    """Test cases for smart HTTP server with side-band-64k and no-done
    support using aiohttp.
    """

    # no-done was introduced in git 1.7.4
    min_git_version = (1, 7, 4)

    def _make_app(self, repo):
        return create_repo_app(repo)


@skipIf(aiohttp_missing, "aiohttp not available")
@skipIf(sys.platform == "win32", "Broken on windows, with very long fail time.")
class DumbAiohttpTestCase(AiohttpServerTests, CompatTestCase):
    """Test cases for dumb HTTP server using aiohttp."""

    def _make_app(self, repo):
        return create_repo_app(repo, dumb=True)

    def test_push_to_dulwich(self) -> NoReturn:
        # Note: remove this if dulwich implements dumb web pushing.
        raise SkipTest("Dumb web pushing not supported.")

    def test_push_to_dulwich_remove_branch(self) -> NoReturn:
        # Note: remove this if dumb pushing is supported
        raise SkipTest("Dumb web pushing not supported.")

    def test_new_shallow_clone_from_dulwich(self) -> NoReturn:
        # Note: remove this if C git and dulwich implement dumb web shallow
        # clones.
        raise SkipTest("Dumb web shallow cloning not supported.")

    def test_shallow_clone_from_git_is_identical(self) -> NoReturn:
        # Note: remove this if C git and dulwich implement dumb web shallow
        # clones.
        raise SkipTest("Dumb web shallow cloning not supported.")

    def test_fetch_same_depth_into_shallow_clone_from_dulwich(self) -> NoReturn:
        # Note: remove this if C git and dulwich implement dumb web shallow
        # clones.
        raise SkipTest("Dumb web shallow cloning not supported.")

    def test_fetch_full_depth_into_shallow_clone_from_dulwich(self) -> NoReturn:
        # Note: remove this if C git and dulwich implement dumb web shallow
        # clones.
        raise SkipTest("Dumb web shallow cloning not supported.")

    def test_push_to_dulwich_issue_88_standard(self) -> NoReturn:
        raise SkipTest("Dumb web pushing not supported.")



# test_midx.py -- Compatibility tests for multi-pack-index functionality
# Copyright (C) 2025 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.

"""Compatibility tests for Git multi-pack-index functionality.

These tests verify that dulwich's MIDX implementation can read and interact
with MIDX files created by C Git, and that Git can read MIDX files created
by Dulwich.
"""

import os
import tempfile

from dulwich.midx import load_midx
from dulwich.object_store import DiskObjectStore
from dulwich.repo import Repo

from .utils import CompatTestCase, run_git_or_fail


class MIDXCompatTests(CompatTestCase):
    """Compatibility tests for multi-pack-index functionality."""

    # Multi-pack-index was introduced in Git 2.21.0
    min_git_version = (2, 21, 0)

    def setUp(self):
        super().setUp()
        self.test_dir = tempfile.mkdtemp()
        self.repo_path = os.path.join(self.test_dir, "test-repo")

        # Set up git identity to avoid committer identity errors
        self.overrideEnv("GIT_COMMITTER_NAME", "Test Author")
        self.overrideEnv("GIT_COMMITTER_EMAIL", "test@example.com")
        self.overrideEnv("GIT_AUTHOR_NAME", "Test Author")
        self.overrideEnv("GIT_AUTHOR_EMAIL", "test@example.com")

    def tearDown(self):
        from .utils import rmtree_ro

        rmtree_ro(self.test_dir)

    def create_test_repo_with_packs(self):
        """Create a test repository with multiple pack files."""
        # Initialize repository
        run_git_or_fail(["init"], cwd=self.test_dir)
        os.rename(os.path.join(self.test_dir, ".git"), self.repo_path)

        work_dir = os.path.join(self.test_dir, "work")
        os.makedirs(work_dir)

        # Create .git file pointing to our repo
        with open(os.path.join(work_dir, ".git"), "w") as f:
            f.write(f"gitdir: {self.repo_path}\n")

        # Create some commits and pack them
        for i in range(5):
            filename = f"file{i}.txt"
            with open(os.path.join(work_dir, filename), "w") as f:
                f.write(f"Content {i}\n" * 100)  # Make files bigger to ensure packing

            run_git_or_fail(["add", filename], cwd=work_dir)
            run_git_or_fail(
                [
                    "commit",
                    "-m",
                    f"Commit {i}",
                    "--author",
                    "Test Author <test@example.com>",
                ],
                cwd=work_dir,
            )

            # Create a pack file after each commit to get multiple packs
            if i > 0:  # Skip first commit to avoid empty pack
                run_git_or_fail(["repack", "-d"], cwd=work_dir)

        return work_dir

    def test_read_git_midx(self):
        """Test that Dulwich can read a MIDX file created by Git."""
        work_dir = self.create_test_repo_with_packs()

        # Have Git create a MIDX file
        run_git_or_fail(["multi-pack-index", "write"], cwd=work_dir)

        # Verify Git created the MIDX file
        midx_path = os.path.join(self.repo_path, "objects", "pack", "multi-pack-index")
        self.assertTrue(
            os.path.exists(midx_path), "Git did not create multi-pack-index file"
        )

        # Load the MIDX file with Dulwich
        midx = load_midx(midx_path)
        try:
            # Verify we can read it
            self.assertGreater(len(midx), 0, "MIDX should contain objects")
            self.assertGreater(midx.pack_count, 0, "MIDX should reference packs")

            # Verify the pack names look reasonable
            # Git stores .idx extensions in MIDX files
            for pack_name in midx.pack_names:
                self.assertTrue(pack_name.startswith("pack-"))
                self.assertTrue(pack_name.endswith(".idx"))
        finally:
            midx.close()

    def test_git_uses_dulwich_midx(self):
        """Test that Git can use a MIDX file created by Dulwich."""
        work_dir = self.create_test_repo_with_packs()

        # Use Dulwich to create a MIDX file
        repo = Repo(self.repo_path)
        try:
            store = repo.object_store
            self.assertIsInstance(store, DiskObjectStore)

            # Write MIDX with Dulwich
            checksum = store.write_midx()
            self.assertEqual(20, len(checksum))
        finally:
            repo.close()

        # Verify the file was created
        midx_path = os.path.join(self.repo_path, "objects", "pack", "multi-pack-index")
        self.assertTrue(os.path.exists(midx_path))

        # Have Git verify the MIDX file (should succeed with return code 0)
        run_git_or_fail(["multi-pack-index", "verify"], cwd=work_dir)

        # Try to use the MIDX with Git commands
        # This should work if the MIDX is valid
        run_git_or_fail(["fsck"], cwd=work_dir)

    def test_midx_object_lookup_matches_git(self):
        """Test that object lookups through MIDX match Git's results."""
        work_dir = self.create_test_repo_with_packs()

        # Have Git create a MIDX file
        run_git_or_fail(["multi-pack-index", "write"], cwd=work_dir)

        # Load with Dulwich
        repo = Repo(self.repo_path)
        try:
            store = repo.object_store

            # Get MIDX
            midx = store.get_midx()
            self.assertIsNotNone(midx, "MIDX should be loaded")

            # Get all objects from Git
            result = run_git_or_fail(["rev-list", "--all", "--objects"], cwd=work_dir)
            object_shas = [
                line.split()[0].encode("ascii")
                for line in result.decode("utf-8").strip().split("\n")
                if line
            ]

            # Verify we can find these objects through the MIDX
            found_count = 0
            for sha_hex in object_shas:
                # Convert hex to binary
                sha_bin = bytes.fromhex(sha_hex.decode("ascii"))

                # Check if it's in the MIDX
                if sha_bin in midx:
                    found_count += 1

                    # Verify we can get the object location
                    result = midx.object_offset(sha_bin)
                    self.assertIsNotNone(result)
                    pack_name, offset = result
                    self.assertIsInstance(pack_name, str)
                    self.assertIsInstance(offset, int)
                    self.assertGreater(offset, 0)

            # We should find at least some objects in the MIDX
            self.assertGreater(
                found_count, 0, "Should find at least some objects in MIDX"
            )
        finally:
            repo.close()

    def test_midx_with_multiple_packs(self):
        """Test MIDX functionality with multiple pack files."""
        work_dir = self.create_test_repo_with_packs()

        # Create multiple pack files explicitly
        run_git_or_fail(["repack"], cwd=work_dir)
        run_git_or_fail(["repack"], cwd=work_dir)

        # Create MIDX with Git
        run_git_or_fail(["multi-pack-index", "write"], cwd=work_dir)

        # Load with Dulwich
        midx_path = os.path.join(self.repo_path, "objects", "pack", "multi-pack-index")
        midx = load_midx(midx_path)
        try:
            # Should have multiple packs
            # (Exact count may vary depending on Git version and repacking)
            self.assertGreaterEqual(midx.pack_count, 1)

            # Verify we can iterate over all entries
            entries = list(midx.iterentries())
            self.assertGreater(len(entries), 0)

            # All entries should have valid structure
            for sha, pack_name, offset in entries:
                self.assertEqual(20, len(sha))  # SHA-1 is 20 bytes
                self.assertIsInstance(pack_name, str)
                # Git stores .idx extensions in MIDX files
                self.assertTrue(pack_name.endswith(".idx"))
                self.assertIsInstance(offset, int)
                self.assertGreaterEqual(offset, 0)
        finally:
            midx.close()

    def test_dulwich_object_store_with_git_midx(self):
        """Test that DiskObjectStore can use Git-created MIDX for lookups."""
        work_dir = self.create_test_repo_with_packs()

        # Have Git create a MIDX file
        run_git_or_fail(["multi-pack-index", "write"], cwd=work_dir)

        # Load repo with Dulwich
        repo = Repo(self.repo_path)
        try:
            # Get a commit from the repo
            result = run_git_or_fail(["rev-parse", "HEAD"], cwd=work_dir)
            head_sha = result.decode("utf-8").strip().encode("ascii")

            # Verify we can access it through Dulwich
            # This should use the MIDX for lookup
            obj = repo.object_store[head_sha]
            self.assertIsNotNone(obj)
            self.assertEqual(b"commit", obj.type_name)
        finally:
            repo.close()

    def test_read_git_loose_objects_pack(self):
        """Dulwich must read packs named "loose-<hash>" written by git.

        ``git maintenance run --task=loose-objects`` packs loose objects into
        a pack named "loose-<hash>.pack" and references it from the MIDX as
        "loose-<hash>.idx". Regression for issue #2229, where the MIDX read
        path asserted every pack name began with "pack-".
        """
        work_dir = self.create_test_repo_with_packs()

        # Add a loose object, then let git pack it into a "loose-<hash>" pack
        # and index it in a MIDX. --no-quiet keeps the pack around.
        with open(os.path.join(work_dir, "loose.txt"), "wb") as f:
            f.write(b"loose object content\n")
        run_git_or_fail(["add", "loose.txt"], cwd=work_dir)
        result = run_git_or_fail(["rev-parse", ":loose.txt"], cwd=work_dir)
        blob_sha = result.decode("utf-8").strip().encode("ascii")

        run_git_or_fail(["maintenance", "run", "--task=loose-objects"], cwd=work_dir)
        # Drop the loose copy so the object lives only in the loose-* pack.
        run_git_or_fail(["prune-packed"], cwd=work_dir)
        run_git_or_fail(["multi-pack-index", "write"], cwd=work_dir)

        pack_dir = os.path.join(self.repo_path, "objects", "pack")
        loose_packs = [
            n
            for n in os.listdir(pack_dir)
            if n.startswith("loose-") and n.endswith(".pack")
        ]
        self.assertNotEqual([], loose_packs, "git did not write a loose-* pack")

        repo = Repo(self.repo_path)
        try:
            obj = repo.object_store[blob_sha]
            self.assertEqual(b"blob", obj.type_name)
            self.assertEqual(b"loose object content\n", obj.data)
        finally:
            repo.close()

    def test_repack_with_midx(self):
        """Test that repacking works correctly with MIDX present."""
        work_dir = self.create_test_repo_with_packs()

        # Create MIDX with Dulwich
        repo = Repo(self.repo_path)
        try:
            repo.object_store.write_midx()
        finally:
            repo.close()

        # Verify Git can still repack
        run_git_or_fail(["repack", "-d"], cwd=work_dir)

        # The MIDX should still be readable
        midx_path = os.path.join(self.repo_path, "objects", "pack", "multi-pack-index")
        if os.path.exists(midx_path):  # Git may remove it during repack
            midx = load_midx(midx_path)
            try:
                self.assertGreaterEqual(len(midx), 0)
            finally:
                midx.close()



# test_objectspec.py -- tests for objectspec.py
# Copyright (C) 2014 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for revision spec parsing."""

# TODO: Round-trip parse-serialize-parse and serialize-parse-serialize tests.

from dulwich.objects import Blob, Commit, Tag, Tree
from dulwich.objectspec import (
    parse_commit,
    parse_commit_range,
    parse_object,
    parse_ref,
    parse_refs,
    parse_reftuple,
    parse_reftuples,
    parse_tree,
)
from dulwich.repo import MemoryRepo
from dulwich.tests.utils import build_commit_graph

from . import TestCase


class ParseObjectTests(TestCase):
    """Test parse_object."""

    def test_nonexistent(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        self.assertRaises(KeyError, parse_object, r, "thisdoesnotexist")

    def test_blob_by_sha(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        b = Blob.from_string(b"Blah")
        r.object_store.add_object(b)
        self.assertEqual(b, parse_object(r, b.id))

    def test_parent_caret(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        c1, c2, c3 = build_commit_graph(r.object_store, [[1], [2, 1], [3, 1, 2]])
        # c3's parents are [c1, c2]
        self.assertEqual(c1, parse_object(r, c3.id + b"^1"))
        self.assertEqual(c1, parse_object(r, c3.id + b"^"))  # ^ defaults to ^1
        self.assertEqual(c2, parse_object(r, c3.id + b"^2"))

    def test_parent_tilde(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        c1, c2, c3 = build_commit_graph(r.object_store, [[1], [2, 1], [3, 2]])
        self.assertEqual(c2, parse_object(r, c3.id + b"~"))
        self.assertEqual(c2, parse_object(r, c3.id + b"~1"))
        self.assertEqual(c1, parse_object(r, c3.id + b"~2"))

    def test_combined_operators(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        c1, c2, _c3, c4 = build_commit_graph(
            r.object_store, [[1], [2, 1], [3, 1, 2], [4, 3]]
        )
        # c4~1^2 means: go back 1 generation from c4 (to c3), then take its 2nd parent
        # c3's parents are [c1, c2], so ^2 is c2
        self.assertEqual(c2, parse_object(r, c4.id + b"~1^2"))
        self.assertEqual(c1, parse_object(r, c4.id + b"~^"))

    def test_with_ref(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        c1, c2, c3 = build_commit_graph(r.object_store, [[1], [2, 1], [3, 2]])
        r.refs[b"refs/heads/master"] = c3.id
        self.assertEqual(c2, parse_object(r, b"master~"))
        self.assertEqual(c1, parse_object(r, b"master~2"))

    def test_caret_zero(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        c1, c2 = build_commit_graph(r.object_store, [[1], [2, 1]])
        # ^0 means the commit itself
        self.assertEqual(c2, parse_object(r, c2.id + b"^0"))
        self.assertEqual(c1, parse_object(r, c2.id + b"~^0"))

    def test_missing_parent(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        c1, c2 = build_commit_graph(r.object_store, [[1], [2, 1]])
        # c2 only has 1 parent, so ^2 should fail
        self.assertRaises(ValueError, parse_object, r, c2.id + b"^2")
        # c1 has no parents, so ~ should fail
        self.assertRaises(ValueError, parse_object, r, c1.id + b"~")

    def test_empty_base(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        self.assertRaises(ValueError, parse_object, r, b"~1")
        self.assertRaises(ValueError, parse_object, r, b"^1")

    def test_non_commit_with_operators(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        b = Blob.from_string(b"Blah")
        r.object_store.add_object(b)
        # Can't apply ~ or ^ to a blob
        self.assertRaises(ValueError, parse_object, r, b.id + b"~1")

    def test_tag_dereference(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        [c1] = build_commit_graph(r.object_store, [[1]])
        # Create an annotated tag
        tag = Tag()
        tag.name = b"v1.0"
        tag.message = b"Test tag"
        tag.tag_time = 1234567890
        tag.tag_timezone = 0
        tag.object = (Commit, c1.id)
        tag.tagger = b"Test Tagger <test@example.com>"
        r.object_store.add_object(tag)
        # ^{} dereferences the tag
        self.assertEqual(c1, parse_object(r, tag.id + b"^{}"))

    def test_nested_tag_dereference(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        [c1] = build_commit_graph(r.object_store, [[1]])
        # Create a tag pointing to a commit
        tag1 = Tag()
        tag1.name = b"v1.0"
        tag1.message = b"Test tag"
        tag1.tag_time = 1234567890
        tag1.tag_timezone = 0
        tag1.object = (Commit, c1.id)
        tag1.tagger = b"Test Tagger <test@example.com>"
        r.object_store.add_object(tag1)

        # Create another tag pointing to the first tag
        tag2 = Tag()
        tag2.name = b"v1.0-release"
        tag2.message = b"Release tag"
        tag2.tag_time = 1234567900
        tag2.tag_timezone = 0
        tag2.object = (Tag, tag1.id)
        tag2.tagger = b"Test Tagger <test@example.com>"
        r.object_store.add_object(tag2)

        # ^{} should recursively dereference to the commit
        self.assertEqual(c1, parse_object(r, tag2.id + b"^{}"))

    def test_path_in_tree(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        # Create a blob
        b = Blob.from_string(b"Test content")

        # Create a commit with the blob in its tree
        [c1] = build_commit_graph(r.object_store, [[1]], trees={1: [(b"test.txt", b)]})

        # HEAD:test.txt should return the blob
        r.refs[b"HEAD"] = c1.id
        result = parse_object(r, b"HEAD:test.txt")
        self.assertEqual(b"Test content", result.data)

    def test_path_in_tree_nested(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        # Create blobs
        b1 = Blob.from_string(b"Content 1")
        b2 = Blob.from_string(b"Content 2")

        # For nested trees, we need to create them manually
        # Create subtree
        subtree = Tree()
        subtree.add(b"file.txt", 0o100644, b1.id)
        r.object_store.add_object(b1)
        r.object_store.add_object(subtree)

        # Create main tree
        main_tree = Tree()
        main_tree.add(b"README", 0o100644, b2.id)
        main_tree.add(b"subdir", 0o040000, subtree.id)
        r.object_store.add_object(b2)
        r.object_store.add_object(main_tree)

        # Create commit with our tree
        c = Commit()
        c.tree = main_tree.id
        c.author = c.committer = b"Test User <test@example.com>"
        c.author_time = c.commit_time = 1234567890
        c.author_timezone = c.commit_timezone = 0
        c.message = b"Test commit"
        r.object_store.add_object(c)

        # Lookup nested path
        result = parse_object(r, c.id + b":subdir/file.txt")
        self.assertEqual(b"Content 1", result.data)

    def test_reflog_lookup(self) -> None:
        # Use a real repo for reflog testing
        import tempfile

        from dulwich.repo import Repo

        with tempfile.TemporaryDirectory() as tmpdir:
            r = Repo.init_bare(tmpdir)
            c1, c2, c3 = build_commit_graph(r.object_store, [[1], [2, 1], [3, 2]])

            # Write reflog entries using the repo's _write_reflog method
            # These are written in chronological order (oldest first)
            r._write_reflog(
                b"HEAD",
                None,
                c1.id,
                b"Test User <test@example.com>",
                1234567890,
                0,
                b"commit: Initial commit",
            )
            r._write_reflog(
                b"HEAD",
                c1.id,
                c2.id,
                b"Test User <test@example.com>",
                1234567891,
                0,
                b"commit: Second commit",
            )
            r._write_reflog(
                b"HEAD",
                c2.id,
                c3.id,
                b"Test User <test@example.com>",
                1234567892,
                0,
                b"commit: Third commit",
            )

            # HEAD@{0} is the most recent (c3)
            self.assertEqual(c3, parse_object(r, b"HEAD@{0}"))
            # HEAD@{1} is the second most recent (c2)
            self.assertEqual(c2, parse_object(r, b"HEAD@{1}"))
            # HEAD@{2} is the third/oldest (c1)
            self.assertEqual(c1, parse_object(r, b"HEAD@{2}"))

    def test_reflog_time_lookup(self) -> None:
        # Use a real repo for reflog testing with time specifications
        import tempfile

        from dulwich.repo import Repo

        with tempfile.TemporaryDirectory() as tmpdir:
            r = Repo.init_bare(tmpdir)
            c1, c2, c3 = build_commit_graph(r.object_store, [[1], [2, 1], [3, 2]])

            # Write reflog entries with specific timestamps
            # 1234567890 = 2009-02-13 23:31:30 UTC
            r._write_reflog(
                b"HEAD",
                None,
                c1.id,
                b"Test User <test@example.com>",
                1234567890,
                0,
                b"commit: Initial commit",
            )
            # 1234657890 = 2009-02-14 23:31:30 UTC (1 day + 1 second later)
            r._write_reflog(
                b"HEAD",
                c1.id,
                c2.id,
                b"Test User <test@example.com>",
                1234657890,
                0,
                b"commit: Second commit",
            )
            # 1235000000 = 2009-02-18 19:33:20 UTC
            r._write_reflog(
                b"HEAD",
                c2.id,
                c3.id,
                b"Test User <test@example.com>",
                1235000000,
                0,
                b"commit: Third commit",
            )

            # Lookup by timestamp - should get the most recent entry at or before time
            self.assertEqual(c1, parse_object(r, b"HEAD@{1234567890}"))
            self.assertEqual(c2, parse_object(r, b"HEAD@{1234657890}"))
            self.assertEqual(c3, parse_object(r, b"HEAD@{1235000000}"))
            # Future timestamp should get latest entry
            self.assertEqual(c3, parse_object(r, b"HEAD@{9999999999}"))

    def test_index_path_lookup_stage0(self) -> None:
        # Test index path lookup for stage 0 (normal files)
        import tempfile

        from dulwich.repo import Repo

        with tempfile.TemporaryDirectory() as tmpdir:
            r = Repo.init(tmpdir)

            # Create a blob and add it to the index
            b = Blob.from_string(b"Test content")
            r.object_store.add_object(b)

            # Add to index
            index = r.open_index()
            from dulwich.index import IndexEntry

            index[b"test.txt"] = IndexEntry(
                ctime=(0, 0),
                mtime=(0, 0),
                dev=0,
                ino=0,
                mode=0o100644,
                uid=0,
                gid=0,
                size=len(b.data),
                sha=b.id,
            )
            index.write()

            # Test :path syntax (defaults to stage 0)
            result = parse_object(r, b":test.txt")
            self.assertEqual(b"Test content", result.data)

            # Test :0:path syntax (explicit stage 0)
            result = parse_object(r, b":0:test.txt")
            self.assertEqual(b"Test content", result.data)

    def test_index_path_lookup_conflicts(self) -> None:
        # Test index path lookup with merge conflicts (stages 1-3)
        import tempfile

        from dulwich.index import ConflictedIndexEntry, IndexEntry
        from dulwich.repo import Repo

        with tempfile.TemporaryDirectory() as tmpdir:
            r = Repo.init(tmpdir)

            # Create three different versions of a file
            b_ancestor = Blob.from_string(b"Ancestor content")
            b_this = Blob.from_string(b"This content")
            b_other = Blob.from_string(b"Other content")
            r.object_store.add_object(b_ancestor)
            r.object_store.add_object(b_this)
            r.object_store.add_object(b_other)

            # Add conflicted entry to index
            index = r.open_index()
            index[b"conflict.txt"] = ConflictedIndexEntry(
                ancestor=IndexEntry(
                    ctime=(0, 0),
                    mtime=(0, 0),
                    dev=0,
                    ino=0,
                    mode=0o100644,
                    uid=0,
                    gid=0,
                    size=len(b_ancestor.data),
                    sha=b_ancestor.id,
                ),
                this=IndexEntry(
                    ctime=(0, 0),
                    mtime=(0, 0),
                    dev=0,
                    ino=0,
                    mode=0o100644,
                    uid=0,
                    gid=0,
                    size=len(b_this.data),
                    sha=b_this.id,
                ),
                other=IndexEntry(
                    ctime=(0, 0),
                    mtime=(0, 0),
                    dev=0,
                    ino=0,
                    mode=0o100644,
                    uid=0,
                    gid=0,
                    size=len(b_other.data),
                    sha=b_other.id,
                ),
            )
            index.write()

            # Test stage 1 (ancestor)
            result = parse_object(r, b":1:conflict.txt")
            self.assertEqual(b"Ancestor content", result.data)

            # Test stage 2 (this)
            result = parse_object(r, b":2:conflict.txt")
            self.assertEqual(b"This content", result.data)

            # Test stage 3 (other)
            result = parse_object(r, b":3:conflict.txt")
            self.assertEqual(b"Other content", result.data)

            # Test that :conflict.txt raises an error for conflicted files
            self.assertRaises(ValueError, parse_object, r, b":conflict.txt")

    def test_index_path_not_found(self) -> None:
        # Test error when path not in index
        import tempfile

        from dulwich.repo import Repo

        with tempfile.TemporaryDirectory() as tmpdir:
            r = Repo.init(tmpdir)

            # Try to lookup non-existent path
            self.assertRaises(KeyError, parse_object, r, b":nonexistent.txt")


class ParseCommitRangeTests(TestCase):
    """Test parse_commit_range."""

    def test_nonexistent(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        self.assertRaises(KeyError, parse_commit_range, r, "thisdoesnotexist..HEAD")

    def test_commit_by_sha(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        c1, _c2, _c3 = build_commit_graph(r.object_store, [[1], [2, 1], [3, 1, 2]])
        self.assertIsNone(parse_commit_range(r, c1.id))

    def test_commit_range(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        c1, c2, _c3 = build_commit_graph(r.object_store, [[1], [2, 1], [3, 1, 2]])
        result = parse_commit_range(r, f"{c1.id.decode()}..{c2.id.decode()}")
        self.assertIsNotNone(result)
        start_commit, end_commit = result
        self.assertEqual(c1, start_commit)
        self.assertEqual(c2, end_commit)


class ParseCommitTests(TestCase):
    """Test parse_commit."""

    def test_nonexistent(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        self.assertRaises(KeyError, parse_commit, r, "thisdoesnotexist")

    def test_commit_by_sha(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        [c1] = build_commit_graph(r.object_store, [[1]])
        self.assertEqual(c1, parse_commit(r, c1.id))

    def test_commit_by_short_sha(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        [c1] = build_commit_graph(r.object_store, [[1]])
        self.assertEqual(c1, parse_commit(r, c1.id[:10]))

    def test_annotated_tag(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        [c1] = build_commit_graph(r.object_store, [[1]])
        # Create an annotated tag pointing to the commit
        tag = Tag()
        tag.name = b"v1.0"
        tag.message = b"Test tag"
        tag.tag_time = 1234567890
        tag.tag_timezone = 0
        tag.object = (Commit, c1.id)
        tag.tagger = b"Test Tagger <test@example.com>"
        r.object_store.add_object(tag)
        # parse_commit should follow the tag to the commit
        self.assertEqual(c1, parse_commit(r, tag.id))

    def test_nested_tags(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        [c1] = build_commit_graph(r.object_store, [[1]])
        # Create an annotated tag pointing to the commit
        tag1 = Tag()
        tag1.name = b"v1.0"
        tag1.message = b"Test tag"
        tag1.tag_time = 1234567890
        tag1.tag_timezone = 0
        tag1.object = (Commit, c1.id)
        tag1.tagger = b"Test Tagger <test@example.com>"
        r.object_store.add_object(tag1)

        # Create another tag pointing to the first tag
        tag2 = Tag()
        tag2.name = b"v1.0-release"
        tag2.message = b"Release tag"
        tag2.tag_time = 1234567900
        tag2.tag_timezone = 0
        tag2.object = (Tag, tag1.id)
        tag2.tagger = b"Test Tagger <test@example.com>"
        r.object_store.add_object(tag2)

        # parse_commit should follow both tags to the commit
        self.assertEqual(c1, parse_commit(r, tag2.id))

    def test_tag_to_missing_commit(self) -> None:
        r = MemoryRepo()
        # Create a tag pointing to a non-existent commit
        missing_sha = b"1234567890123456789012345678901234567890"
        tag = Tag()
        tag.name = b"v1.0"
        tag.message = b"Test tag"
        tag.tag_time = 1234567890
        tag.tag_timezone = 0
        tag.object = (Commit, missing_sha)
        tag.tagger = b"Test Tagger <test@example.com>"
        r.object_store.add_object(tag)

        # Should raise KeyError for missing commit
        self.assertRaises(KeyError, parse_commit, r, tag.id)

    def test_tag_to_blob(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        # Create a blob
        blob = Blob.from_string(b"Test content")
        r.object_store.add_object(blob)

        # Create a tag pointing to the blob
        tag = Tag()
        tag.name = b"blob-tag"
        tag.message = b"Tag pointing to blob"
        tag.tag_time = 1234567890
        tag.tag_timezone = 0
        tag.object = (Blob, blob.id)
        tag.tagger = b"Test Tagger <test@example.com>"
        r.object_store.add_object(tag)

        # Should raise ValueError as it's not a commit
        self.assertRaises(ValueError, parse_commit, r, tag.id)

    def test_commit_object(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        [c1] = build_commit_graph(r.object_store, [[1]])
        # Test that passing a Commit object directly returns the same object
        self.assertEqual(c1, parse_commit(r, c1))


class ParseRefTests(TestCase):
    def test_nonexistent(self) -> None:
        r = {}
        self.assertRaises(KeyError, parse_ref, r, b"thisdoesnotexist")

    def test_ambiguous_ref(self) -> None:
        r = {
            b"ambig1": "bla",
            b"refs/ambig1": "bla",
            b"refs/tags/ambig1": "bla",
            b"refs/heads/ambig1": "bla",
            b"refs/remotes/ambig1": "bla",
            b"refs/remotes/ambig1/HEAD": "bla",
        }
        self.assertEqual(b"ambig1", parse_ref(r, b"ambig1"))

    def test_ambiguous_ref2(self) -> None:
        r = {
            b"refs/ambig2": "bla",
            b"refs/tags/ambig2": "bla",
            b"refs/heads/ambig2": "bla",
            b"refs/remotes/ambig2": "bla",
            b"refs/remotes/ambig2/HEAD": "bla",
        }
        self.assertEqual(b"refs/ambig2", parse_ref(r, b"ambig2"))

    def test_ambiguous_tag(self) -> None:
        r = {
            b"refs/tags/ambig3": "bla",
            b"refs/heads/ambig3": "bla",
            b"refs/remotes/ambig3": "bla",
            b"refs/remotes/ambig3/HEAD": "bla",
        }
        self.assertEqual(b"refs/tags/ambig3", parse_ref(r, b"ambig3"))

    def test_ambiguous_head(self) -> None:
        r = {
            b"refs/heads/ambig4": "bla",
            b"refs/remotes/ambig4": "bla",
            b"refs/remotes/ambig4/HEAD": "bla",
        }
        self.assertEqual(b"refs/heads/ambig4", parse_ref(r, b"ambig4"))

    def test_ambiguous_remote(self) -> None:
        r = {b"refs/remotes/ambig5": "bla", b"refs/remotes/ambig5/HEAD": "bla"}
        self.assertEqual(b"refs/remotes/ambig5", parse_ref(r, b"ambig5"))

    def test_ambiguous_remote_head(self) -> None:
        r = {b"refs/remotes/ambig6/HEAD": "bla"}
        self.assertEqual(b"refs/remotes/ambig6/HEAD", parse_ref(r, b"ambig6"))

    def test_heads_full(self) -> None:
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual(b"refs/heads/foo", parse_ref(r, b"refs/heads/foo"))

    def test_heads_partial(self) -> None:
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual(b"refs/heads/foo", parse_ref(r, b"heads/foo"))

    def test_tags_partial(self) -> None:
        r = {b"refs/tags/foo": "bla"}
        self.assertEqual(b"refs/tags/foo", parse_ref(r, b"tags/foo"))


class ParseRefsTests(TestCase):
    def test_nonexistent(self) -> None:
        r = {}
        self.assertRaises(KeyError, parse_refs, r, [b"thisdoesnotexist"])

    def test_head(self) -> None:
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual([b"refs/heads/foo"], parse_refs(r, [b"foo"]))

    def test_full(self) -> None:
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual([b"refs/heads/foo"], parse_refs(r, b"refs/heads/foo"))


class ParseReftupleTests(TestCase):
    def test_nonexistent(self) -> None:
        r = {}
        self.assertRaises(KeyError, parse_reftuple, r, r, b"thisdoesnotexist")

    def test_head(self) -> None:
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual(
            (b"refs/heads/foo", b"refs/heads/foo", False),
            parse_reftuple(r, r, b"foo"),
        )
        self.assertEqual(
            (b"refs/heads/foo", b"refs/heads/foo", True),
            parse_reftuple(r, r, b"+foo"),
        )
        self.assertEqual(
            (b"refs/heads/foo", b"refs/heads/foo", True),
            parse_reftuple(r, {}, b"+foo"),
        )
        self.assertEqual(
            (b"refs/heads/foo", b"refs/heads/foo", True),
            parse_reftuple(r, {}, b"foo", True),
        )

    def test_full(self) -> None:
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual(
            (b"refs/heads/foo", b"refs/heads/foo", False),
            parse_reftuple(r, r, b"refs/heads/foo"),
        )

    def test_no_left_ref(self) -> None:
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual(
            (None, b"refs/heads/foo", False),
            parse_reftuple(r, r, b":refs/heads/foo"),
        )

    def test_no_right_ref(self) -> None:
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual(
            (b"refs/heads/foo", None, False),
            parse_reftuple(r, r, b"refs/heads/foo:"),
        )

    def test_default_with_string(self) -> None:
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual(
            (b"refs/heads/foo", b"refs/heads/foo", False),
            parse_reftuple(r, r, "foo"),
        )


class ParseReftuplesTests(TestCase):
    def test_nonexistent(self) -> None:
        r = {}
        self.assertRaises(KeyError, parse_reftuples, r, r, [b"thisdoesnotexist"])

    def test_head(self) -> None:
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual(
            [(b"refs/heads/foo", b"refs/heads/foo", False)],
            parse_reftuples(r, r, [b"foo"]),
        )

    def test_full(self) -> None:
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual(
            [(b"refs/heads/foo", b"refs/heads/foo", False)],
            parse_reftuples(r, r, b"refs/heads/foo"),
        )
        r = {b"refs/heads/foo": "bla"}
        self.assertEqual(
            [(b"refs/heads/foo", b"refs/heads/foo", True)],
            parse_reftuples(r, r, b"refs/heads/foo", True),
        )


class ParseTreeTests(TestCase):
    """Test parse_tree."""

    def test_nonexistent(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        self.assertRaises(KeyError, parse_tree, r, "thisdoesnotexist")

    def test_from_commit(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        c1, _c2, _c3 = build_commit_graph(r.object_store, [[1], [2, 1], [3, 1, 2]])
        self.assertEqual(r[c1.tree], parse_tree(r, c1.id))
        self.assertEqual(r[c1.tree], parse_tree(r, c1.tree))

    def test_from_ref(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        c1, _c2, _c3 = build_commit_graph(r.object_store, [[1], [2, 1], [3, 1, 2]])
        r.refs[b"refs/heads/foo"] = c1.id
        self.assertEqual(r[c1.tree], parse_tree(r, b"foo"))

    def test_tree_object(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        [c1] = build_commit_graph(r.object_store, [[1]])
        tree = r[c1.tree]
        # Test that passing a Tree object directly returns the same object
        self.assertEqual(tree, parse_tree(r, tree))

    def test_commit_object(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        [c1] = build_commit_graph(r.object_store, [[1]])
        # Test that passing a Commit object returns its tree
        self.assertEqual(r[c1.tree], parse_tree(r, c1))

    def test_tag_object(self) -> None:
        r = MemoryRepo()
        self.addCleanup(r.close)
        [c1] = build_commit_graph(r.object_store, [[1]])
        # Create an annotated tag pointing to the commit
        tag = Tag()
        tag.name = b"v1.0"
        tag.message = b"Test tag"
        tag.tag_time = 1234567890
        tag.tag_timezone = 0
        tag.object = (Commit, c1.id)
        tag.tagger = b"Test Tagger <test@example.com>"
        r.object_store.add_object(tag)
        # parse_tree should follow the tag to the commit's tree
        self.assertEqual(r[c1.tree], parse_tree(r, tag))



# test_rebase.py -- tests for porcelain rebase
# Copyright (C) 2025 Dulwich contributors
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for porcelain rebase functions."""

import os
import tempfile

from dulwich import porcelain
from dulwich.repo import Repo

from .. import TestCase


class RebasePorcelainTestCase(TestCase):
    """Tests for the porcelain rebase function."""

    def setUp(self):
        """Set up test repository."""
        super().setUp()
        self.test_dir = tempfile.mkdtemp()
        self.repo = Repo.init(self.test_dir)

        # Create initial commit
        with open(os.path.join(self.test_dir, "README.md"), "wb") as f:
            f.write(b"# Test Repository\n")

        self.repo.get_worktree().stage(["README.md"])
        self.initial_commit = self.repo.get_worktree().commit(
            message=b"Initial commit",
            committer=b"Test User <test@example.com>",
            author=b"Test User <test@example.com>",
        )

    def tearDown(self):
        """Clean up test directory."""
        import shutil

        shutil.rmtree(self.test_dir)

    def test_porcelain_rebase(self):
        """Test rebase through porcelain interface."""
        # Create and checkout feature branch
        self.repo.refs[b"refs/heads/feature"] = self.initial_commit
        porcelain.checkout(self.repo, "feature")

        # Add commit to feature branch
        with open(os.path.join(self.test_dir, "feature.txt"), "wb") as f:
            f.write(b"Feature file\n")

        porcelain.add(self.repo, ["feature.txt"])
        porcelain.commit(
            self.repo,
            message="Add feature",
            author="Test User <test@example.com>",
            committer="Test User <test@example.com>",
        )

        # Switch to main and add different commit
        porcelain.checkout(self.repo, "master")

        with open(os.path.join(self.test_dir, "main.txt"), "wb") as f:
            f.write(b"Main file\n")

        porcelain.add(self.repo, ["main.txt"])
        porcelain.commit(
            self.repo,
            message="Main update",
            author="Test User <test@example.com>",
            committer="Test User <test@example.com>",
        )

        # Switch back to feature and rebase
        porcelain.checkout(self.repo, "feature")

        # Perform rebase
        new_shas = porcelain.rebase(self.repo, "master")

        # Should have rebased one commit
        self.assertEqual(len(new_shas), 1)

        # Check that the rebased commit has the correct parent and tree
        feature_head = self.repo.refs[b"refs/heads/feature"]
        feature_commit_obj = self.repo[feature_head]

        # Should have master as parent
        master_head = self.repo.refs[b"refs/heads/master"]
        self.assertEqual(feature_commit_obj.parents, [master_head])

        # Tree should have both files
        tree = self.repo[feature_commit_obj.tree]
        self.assertIn(b"feature.txt", tree)
        self.assertIn(b"main.txt", tree)
        self.assertIn(b"README.md", tree)



# test_bisect.py -- Tests for bisect functionality
# Copyright (C) 2025 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for bisect functionality."""

import os
import shutil
import tempfile

from dulwich.bisect import BisectState
from dulwich.repo import Repo
from dulwich.tests.utils import make_commit

from . import TestCase


class BisectStateTests(TestCase):
    """Tests for BisectState class."""

    def setUp(self) -> None:
        super().setUp()
        self.test_dir = tempfile.mkdtemp()
        self.repo = Repo.init(self.test_dir)

    def tearDown(self) -> None:
        shutil.rmtree(self.test_dir)

    def test_is_active_false(self) -> None:
        """Test is_active when no bisect session is active."""
        state = BisectState(self.repo)
        self.assertFalse(state.is_active)

    def test_start_bisect(self) -> None:
        """Test starting a bisect session."""
        # Create at least one commit so HEAD exists
        c1 = make_commit(id=b"1" * 40, message=b"initial commit")
        self.repo.object_store.add_object(c1)
        self.repo.refs[b"HEAD"] = c1.id
        self.repo.refs[b"refs/heads/main"] = c1.id

        state = BisectState(self.repo)
        state.start()

        self.assertTrue(state.is_active)
        self.assertTrue(
            os.path.exists(os.path.join(self.repo.controldir(), "BISECT_START"))
        )
        self.assertTrue(
            os.path.exists(os.path.join(self.repo.controldir(), "BISECT_TERMS"))
        )
        self.assertTrue(
            os.path.exists(os.path.join(self.repo.controldir(), "BISECT_NAMES"))
        )
        self.assertTrue(
            os.path.exists(os.path.join(self.repo.controldir(), "BISECT_LOG"))
        )

    def test_start_bisect_no_head(self) -> None:
        """Test starting a bisect session when repository has no HEAD."""
        state = BisectState(self.repo)

        with self.assertRaises(ValueError) as cm:
            state.start()
        self.assertIn("Cannot start bisect: repository has no HEAD", str(cm.exception))

    def test_start_bisect_already_active(self) -> None:
        """Test starting a bisect session when one is already active."""
        # Create at least one commit so HEAD exists
        c1 = make_commit(id=b"1" * 40, message=b"initial commit")
        self.repo.object_store.add_object(c1)
        self.repo.refs[b"HEAD"] = c1.id

        state = BisectState(self.repo)
        state.start()

        with self.assertRaises(ValueError):
            state.start()

    def test_mark_bad_no_session(self) -> None:
        """Test marking bad commit when no session is active."""
        state = BisectState(self.repo)

        with self.assertRaises(ValueError):
            state.mark_bad()

    def test_mark_good_no_session(self) -> None:
        """Test marking good commit when no session is active."""
        state = BisectState(self.repo)

        with self.assertRaises(ValueError):
            state.mark_good()

    def test_reset_no_session(self) -> None:
        """Test resetting when no session is active."""
        state = BisectState(self.repo)

        with self.assertRaises(ValueError):
            state.reset()

    def test_bisect_workflow(self) -> None:
        """Test a complete bisect workflow."""
        # Create some commits
        c1 = make_commit(message=b"good commit 1")
        c2 = make_commit(message=b"good commit 2", parents=[c1.id])
        c3 = make_commit(message=b"bad commit", parents=[c2.id])
        c4 = make_commit(message=b"bad commit 2", parents=[c3.id])

        # Add commits to object store
        for commit in [c1, c2, c3, c4]:
            self.repo.object_store.add_object(commit)

        # Set HEAD to latest commit
        self.repo.refs[b"HEAD"] = c4.id

        # Start bisect
        state = BisectState(self.repo)
        state.start()

        # Mark bad and good
        state.mark_bad(c4.id)
        state.mark_good(c1.id)

        # Check that refs were created
        self.assertTrue(
            os.path.exists(
                os.path.join(self.repo.controldir(), "refs", "bisect", "bad")
            )
        )
        self.assertTrue(
            os.path.exists(
                os.path.join(
                    self.repo.controldir(),
                    "refs",
                    "bisect",
                    f"good-{c1.id.decode('ascii')}",
                )
            )
        )

        # Reset
        state.reset()
        self.assertFalse(state.is_active)



# errors.py -- errors for dulwich
# Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
# Copyright (C) 2009-2012 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Dulwich-related exception classes and utility functions."""

__all__ = [
    "ApplyDeltaError",
    "ChecksumMismatch",
    "CommitError",
    "FileFormatException",
    "GitProtocolError",
    "HangupException",
    "HookError",
    "MissingCommitError",
    "NoIndexPresent",
    "NotBlobError",
    "NotCommitError",
    "NotGitRepository",
    "NotTagError",
    "NotTreeError",
    "ObjectFormatException",
    "ObjectMissing",
    "PackedRefsException",
    "RefFormatError",
    "SendPackError",
    "UnexpectedCommandError",
    "WorkingTreeModifiedError",
    "WrongObjectException",
]

# Please do not add more errors here, but instead add them close to the code
# that raises the error.

import binascii
from collections.abc import Sequence


class ChecksumMismatch(Exception):
    """A checksum didn't match the expected contents."""

    def __init__(
        self,
        expected: bytes | str,
        got: bytes | str,
        extra: str | None = None,
    ) -> None:
        """Initialize a ChecksumMismatch exception.

        Args:
            expected: The expected checksum value (bytes or hex string).
            got: The actual checksum value (bytes or hex string).
            extra: Optional additional error information.
        """
        if isinstance(expected, bytes) and len(expected) in (20, 32):
            expected_str = binascii.hexlify(expected).decode("ascii")
        else:
            expected_str = (
                expected if isinstance(expected, str) else expected.decode("ascii")
            )
        if isinstance(got, bytes) and len(got) in (20, 32):
            got_str = binascii.hexlify(got).decode("ascii")
        else:
            got_str = got if isinstance(got, str) else got.decode("ascii")
        self.expected = expected_str
        self.got = got_str
        self.extra = extra
        message = f"Checksum mismatch: Expected {expected_str}, got {got_str}"
        if self.extra is not None:
            message += f"; {extra}"
        Exception.__init__(self, message)


class WrongObjectException(Exception):
    """Baseclass for all the _ is not a _ exceptions on objects.

    Do not instantiate directly.

    Subclasses should define a type_name attribute that indicates what
    was expected if they were raised.
    """

    type_name: str

    def __init__(self, sha: bytes, *args: object, **kwargs: object) -> None:
        """Initialize a WrongObjectException.

        Args:
            sha: The SHA of the object that was not of the expected type.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        Exception.__init__(self, f"{sha.decode('ascii')} is not a {self.type_name}")


class NotCommitError(WrongObjectException):
    """Indicates that the sha requested does not point to a commit."""

    type_name = "commit"


class NotTreeError(WrongObjectException):
    """Indicates that the sha requested does not point to a tree."""

    type_name = "tree"


class NotTagError(WrongObjectException):
    """Indicates that the sha requested does not point to a tag."""

    type_name = "tag"


class NotBlobError(WrongObjectException):
    """Indicates that the sha requested does not point to a blob."""

    type_name = "blob"


class MissingCommitError(Exception):
    """Indicates that a commit was not found in the repository."""

    def __init__(self, sha: bytes, *args: object, **kwargs: object) -> None:
        """Initialize a MissingCommitError.

        Args:
            sha: The SHA of the missing commit.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        self.sha = sha
        Exception.__init__(self, f"{sha.decode('ascii')} is not in the revision store")


class ObjectMissing(Exception):
    """Indicates that a requested object is missing."""

    def __init__(self, sha: bytes, *args: object, **kwargs: object) -> None:
        """Initialize an ObjectMissing exception.

        Args:
            sha: The SHA of the missing object.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        Exception.__init__(self, f"{sha.decode('ascii')} is not in the pack")


class ApplyDeltaError(Exception):
    """Indicates that applying a delta failed."""

    def __init__(self, *args: object, **kwargs: object) -> None:
        """Initialize an ApplyDeltaError.

        Args:
            *args: Error message and additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        Exception.__init__(self, *args, **kwargs)


class NotGitRepository(Exception):
    """Indicates that no Git repository was found."""

    def __init__(self, *args: object, **kwargs: object) -> None:
        """Initialize a NotGitRepository exception.

        Args:
            *args: Error message and additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        Exception.__init__(self, *args, **kwargs)


class GitProtocolError(Exception):
    """Git protocol exception."""

    def __init__(self, *args: object, **kwargs: object) -> None:
        """Initialize a GitProtocolError.

        Args:
            *args: Error message and additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        Exception.__init__(self, *args, **kwargs)

    def __eq__(self, other: object) -> bool:
        """Check equality between GitProtocolError instances.

        Args:
            other: The object to compare with.

        Returns:
            True if both are GitProtocolError instances with same args, False otherwise.
        """
        return isinstance(other, GitProtocolError) and self.args == other.args


class SendPackError(GitProtocolError):
    """An error occurred during send_pack."""


class HangupException(GitProtocolError):
    """Hangup exception."""

    def __init__(self, stderr_lines: Sequence[bytes] | None = None) -> None:
        """Initialize a HangupException.

        Args:
            stderr_lines: Optional list of stderr output lines from the remote server.
        """
        if stderr_lines:
            super().__init__(
                "\n".join(
                    line.decode("utf-8", "surrogateescape") for line in stderr_lines
                )
            )
        else:
            super().__init__("The remote server unexpectedly closed the connection.")
        self.stderr_lines = stderr_lines

    def __eq__(self, other: object) -> bool:
        """Check equality between HangupException instances.

        Args:
            other: The object to compare with.

        Returns:
            True if both are HangupException instances with same stderr_lines, False otherwise.
        """
        return (
            isinstance(other, HangupException)
            and self.stderr_lines == other.stderr_lines
        )


class UnexpectedCommandError(GitProtocolError):
    """Unexpected command received in a proto line."""

    def __init__(self, command: str | None) -> None:
        """Initialize an UnexpectedCommandError.

        Args:
            command: The unexpected command received, or None for flush-pkt.
        """
        command_str = "flush-pkt" if command is None else f"command {command}"
        super().__init__(f"Protocol got unexpected {command_str}")


class FileFormatException(Exception):
    """Base class for exceptions relating to reading git file formats."""


class PackedRefsException(FileFormatException):
    """Indicates an error parsing a packed-refs file."""


class ObjectFormatException(FileFormatException):
    """Indicates an error parsing an object."""


class NoIndexPresent(Exception):
    """No index is present."""


class CommitError(Exception):
    """An error occurred while performing a commit."""


class RefFormatError(Exception):
    """Indicates an invalid ref name."""


class HookError(Exception):
    """An error occurred while executing a hook."""


class WorkingTreeModifiedError(Exception):
    """Indicates that the working tree has modifications that would be overwritten."""



# subtree.py -- Porcelain for subtree operations
# Copyright (C) 2025 Dulwich contributors
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Porcelain for git subtree operations."""

__all__ = [
    "subtree_add",
    "subtree_merge",
    "subtree_pull",
    "subtree_push",
    "subtree_split",
]

import os
import time
from typing import TYPE_CHECKING

from dulwich.errors import NotCommitError
from dulwich.graph import find_merge_base
from dulwich.merge import three_way_merge
from dulwich.objects import Commit, ObjectID, Tag, Tree
from dulwich.objectspec import parse_commit
from dulwich.refs import LOCAL_BRANCH_PREFIX
from dulwich.repo import Repo
from dulwich.subtree import (
    add_subtree_metadata,
    create_tree_with_subtree,
    extract_subtree,
)

if TYPE_CHECKING:
    from . import RepoPath
else:
    RepoPath = str | os.PathLike[str] | Repo


def _create_subtree_commit(
    repo: Repo,
    tree_id: ObjectID,
    parents: list[ObjectID],
    message: bytes,
    base_commit: Commit,
) -> ObjectID:
    """Create a subtree commit with standard metadata.

    Args:
      repo: Repository to create commit in
      tree_id: Tree ID for the commit
      parents: Parent commit IDs
      message: Commit message (with trailers already added)
      base_commit: Base commit to copy author/committer info from

    Returns:
      Object ID of the new commit
    """
    new_commit_obj = base_commit.copy()
    if not isinstance(new_commit_obj, Commit):
        raise TypeError("Expected Commit copy")
    new_commit_obj.tree = tree_id
    new_commit_obj.parents = parents
    new_commit_obj.author_time = int(time.time())
    new_commit_obj.commit_time = int(time.time())
    new_commit_obj.message = message

    repo.object_store.add_object(new_commit_obj)
    return new_commit_obj.id


def subtree_add(
    repo: RepoPath,
    prefix: bytes | str,
    repository: str | None = None,
    ref: bytes | str | None = None,
    commit: bytes | str | Commit | Tag | None = None,
    squash: bool = False,
    message: bytes | None = None,
) -> ObjectID:
    """Add a subtree to the repository.

    Args:
      repo: Repository to add subtree to
      prefix: Directory path where subtree will be added
      repository: URL of repository to add (for remote subtrees)
      ref: Ref to fetch from repository
      commit: Commit object to add (for local subtrees)
      squash: If True, squash history into a single commit
      message: Custom commit message

    Returns:
      Object ID of the new commit

    Raises:
      Error: If prefix already exists or commit cannot be found
      ValueError: If neither (repository, ref) nor commit is provided
    """
    from . import Error, fetch, open_repo_closing

    if isinstance(prefix, str):
        prefix = prefix.encode("utf-8")

    with open_repo_closing(repo) as r:
        # Validate that prefix doesn't already exist
        head_commit = r[b"HEAD"]
        if not isinstance(head_commit, Commit):
            raise Error("HEAD is not a commit")

        head_tree = r[head_commit.tree]
        if not isinstance(head_tree, Tree):
            raise Error("HEAD tree is invalid")

        if extract_subtree(r.object_store, head_tree, prefix) is not None:
            raise Error(f"prefix '{prefix.decode('utf-8', 'replace')}' already exists")

        # Fetch or parse the commit to add
        if commit is not None:
            # Local commit by SHA or commit object already in this repo
            if isinstance(commit, bytes | str):
                if isinstance(commit, str):
                    commit = commit.encode("utf-8")
                subtree_commit = parse_commit(r, commit)
            elif isinstance(commit, Tag):
                obj = r[commit.object[1]]
                if not isinstance(obj, Commit):
                    raise NotCommitError(commit.object[1])
                subtree_commit = obj
            else:
                # Commit object - must already be in this repo's object store
                subtree_commit = commit
        elif repository and ref:
            # Remote repository - fetch it
            if isinstance(ref, str):
                ref = ref.encode("utf-8")

            # Fetch from remote (fetches all refs)
            from io import StringIO

            fetch_result = fetch(r, repository, outstream=StringIO())

            # Get the fetched commit
            from dulwich.refs import Ref

            if ref.startswith(b"refs/"):
                ref_key = Ref(ref)
            else:
                ref_key = Ref(LOCAL_BRANCH_PREFIX + ref)

            commit_id = fetch_result.refs.get(ref_key)
            if commit_id is None:
                # Ref not found, maybe it's a commit SHA already in the repo
                try:
                    subtree_commit_obj = r[ref]
                    if not isinstance(subtree_commit_obj, Commit):
                        raise NotCommitError(ref)
                    subtree_commit = subtree_commit_obj
                except KeyError:
                    raise Error(
                        f"Ref or commit '{ref.decode('utf-8', 'replace')}' not found"
                    )
            else:
                subtree_commit_obj = r[commit_id]
                if not isinstance(subtree_commit_obj, Commit):
                    raise NotCommitError(commit_id)
                subtree_commit = subtree_commit_obj
        else:
            raise ValueError("Either commit or (repository, ref) must be provided")

        # Get the tree to add
        subtree_tree = r[subtree_commit.tree]
        if not isinstance(subtree_tree, Tree):
            raise Error("Subtree commit tree is invalid")

        # Create new tree with subtree inserted
        new_tree_id = create_tree_with_subtree(
            r.object_store, head_tree, prefix, subtree_tree
        )

        # Create commit message
        if message is None:
            if repository:
                ref_str = (
                    ref.decode("utf-8", "replace") if isinstance(ref, bytes) else ref
                )
                subtree_desc = f"'{repository}' at '{ref_str}'"
            else:
                subtree_desc = str(subtree_commit.id)
            message = f"Add '{prefix.decode('utf-8', 'replace')}' from {subtree_desc}".encode()

        # Add subtree metadata
        message = add_subtree_metadata(message, prefix, split=subtree_commit.id)

        # Create commit parents
        parents = [head_commit.id]
        if not squash:
            parents.append(subtree_commit.id)

        # Create new commit
        new_commit_id = _create_subtree_commit(
            r, new_tree_id, parents, message, head_commit
        )

        # Update HEAD
        r[b"HEAD"] = new_commit_id

        return new_commit_id


def subtree_merge(
    repo: RepoPath,
    prefix: bytes | str,
    commit: bytes | str | Commit | Tag,
    squash: bool = False,
    message: bytes | None = None,
) -> ObjectID:
    """Merge changes into an existing subtree.

    Args:
      repo: Repository to merge into
      prefix: Directory path of the subtree
      commit: Commit to merge
      squash: If True, squash history into a single commit
      message: Custom commit message

    Returns:
      Object ID of the merge commit

    Raises:
      Error: If prefix doesn't exist or merge fails
    """
    from . import Error, open_repo_closing

    if isinstance(prefix, str):
        prefix = prefix.encode("utf-8")

    with open_repo_closing(repo) as r:
        # Parse the commit to merge
        if isinstance(commit, bytes | str):
            if isinstance(commit, str):
                commit = commit.encode("utf-8")
            merge_commit = parse_commit(r, commit)
        elif isinstance(commit, Tag):
            obj = r[commit.object[1]]
            if not isinstance(obj, Commit):
                raise NotCommitError(commit.object[1])
            merge_commit = obj
        else:
            merge_commit = commit

        # Get current HEAD
        head_commit = r[b"HEAD"]
        if not isinstance(head_commit, Commit):
            raise Error("HEAD is not a commit")

        # Validate that prefix exists
        head_tree = r[head_commit.tree]
        if not isinstance(head_tree, Tree):
            raise Error("HEAD tree is invalid")

        if extract_subtree(r.object_store, head_tree, prefix) is None:
            raise Error(f"prefix '{prefix.decode('utf-8', 'replace')}' does not exist")

        # Find the merge base for proper three-way merge
        merge_bases = find_merge_base(r, [head_commit.id, merge_commit.id])

        if merge_bases:
            # Get the base commit for the merge
            base_commit_obj = r[merge_bases[0]]
            if not isinstance(base_commit_obj, Commit):
                raise Error("Merge base is not a commit")
        else:
            # No common ancestor - this is unusual for subtree merge
            base_commit_obj = None

        # Perform three-way merge on the subtree portion
        merge_tree_obj = r[merge_commit.tree]
        if not isinstance(merge_tree_obj, Tree):
            raise Error("Merge commit tree is invalid")

        # Extract the subtree from each version
        merge_subtree = extract_subtree(r.object_store, merge_tree_obj, prefix)
        if merge_subtree is None:
            # The merge commit doesn't have this subtree path - treat as deletion
            # For subtree merge, we typically want the whole tree from merge_commit
            merge_subtree = merge_tree_obj

        head_subtree = extract_subtree(r.object_store, head_tree, prefix)
        assert head_subtree is not None  # We already checked this exists

        base_subtree = None
        if base_commit_obj:
            base_tree_obj = r[base_commit_obj.tree]
            if isinstance(base_tree_obj, Tree):
                base_subtree = extract_subtree(r.object_store, base_tree_obj, prefix)

        # Perform three-way merge on the subtrees
        # Create temporary commits to use the merge infrastructure
        temp_head = head_commit.copy()
        assert isinstance(temp_head, Commit)
        temp_head.tree = head_subtree.id

        temp_merge = merge_commit.copy()
        assert isinstance(temp_merge, Commit)
        temp_merge.tree = merge_subtree.id

        if base_subtree and base_commit_obj:
            temp_base_obj = base_commit_obj.copy()
            if not isinstance(temp_base_obj, Commit):
                raise TypeError("Expected Commit copy")
            temp_base_obj.tree = base_subtree.id
            temp_base = temp_base_obj
        else:
            temp_base = None

        # Perform the three-way merge
        merged_subtree, conflicts = three_way_merge(
            r.object_store, temp_base, temp_head, temp_merge
        )

        if conflicts and not squash:
            # Report conflicts but continue - user will need to resolve
            import sys

            for path in conflicts:
                prefix_str = prefix.decode("utf-8", "replace")
                path_str = path.decode("utf-8", "replace")
                print(
                    f"CONFLICT (content): Merge conflict in {prefix_str}/{path_str}",
                    file=sys.stderr,
                )

        # Store the merged subtree
        r.object_store.add_object(merged_subtree)

        # Insert the merged subtree back into the main tree
        new_tree_id = create_tree_with_subtree(
            r.object_store, head_tree, prefix, merged_subtree
        )

        # Create commit message
        if message is None:
            merge_id_str = merge_commit.id.decode("utf-8", "replace")
            prefix_str = prefix.decode("utf-8", "replace")
            message = f"Merge commit '{merge_id_str}' as '{prefix_str}'".encode()

        # Add subtree metadata
        message = add_subtree_metadata(message, prefix, split=merge_commit.id)

        # Create commit parents
        parents = [head_commit.id]
        if not squash:
            parents.append(merge_commit.id)

        # Create merge commit
        new_commit_id = _create_subtree_commit(
            r, new_tree_id, parents, message, head_commit
        )

        # Update HEAD
        r[b"HEAD"] = new_commit_id

        return new_commit_id


def subtree_split(
    repo: RepoPath,
    prefix: bytes | str,
    branch: bytes | str | None = None,
    rejoin: bool = False,
    onto: bytes | str | None = None,
) -> ObjectID:
    """Split a subtree into a separate branch.

    Args:
      repo: Repository to split from
      prefix: Directory path of the subtree to split
      branch: Branch name to create (optional)
      rejoin: If True, merge the split branch back with metadata
      onto: Commit to use as the initial parent (optional)

    Returns:
      Object ID of the split commit

    Raises:
      Error: If prefix doesn't exist
    """
    from . import Error, open_repo_closing

    if isinstance(prefix, str):
        prefix = prefix.encode("utf-8")
    if isinstance(branch, str):
        branch = branch.encode("utf-8")
    if isinstance(onto, str):
        onto = onto.encode("utf-8")

    with open_repo_closing(repo) as r:
        # Get current HEAD
        head_commit = r[b"HEAD"]
        if not isinstance(head_commit, Commit):
            raise Error("HEAD is not a commit")

        # Validate that prefix exists
        head_tree = r[head_commit.tree]
        if not isinstance(head_tree, Tree):
            raise Error("HEAD tree is invalid")

        subtree = extract_subtree(r.object_store, head_tree, prefix)
        if subtree is None:
            raise Error(f"prefix '{prefix.decode('utf-8', 'replace')}' does not exist")

        # Walk the history and extract subtree commits
        def process_commit(
            commit_id: ObjectID,
            cache: dict[ObjectID, ObjectID | None],
        ) -> ObjectID | None:
            """Process a commit and return the rewritten commit ID."""
            if commit_id in cache:
                return cache[commit_id]

            commit = r[commit_id]
            if not isinstance(commit, Commit):
                cache[commit_id] = None
                return None

            # Get the subtree for this commit
            commit_tree = r[commit.tree]
            if not isinstance(commit_tree, Tree):
                cache[commit_id] = None
                return None

            commit_subtree = extract_subtree(r.object_store, commit_tree, prefix)
            if commit_subtree is None:
                # This commit doesn't have the subtree
                cache[commit_id] = None
                return None

            # Process parents
            new_parents: list[ObjectID] = []
            for parent_id in commit.parents:
                parent_split = process_commit(parent_id, cache)
                if parent_split is not None:
                    new_parents.append(parent_split)

            # Add onto commit as parent if specified and this is the first commit
            if onto and not new_parents:
                if isinstance(onto, bytes):
                    onto_commit = parse_commit(r, onto)
                    new_parents.append(onto_commit.id)

            # Check if the subtree changed from parent
            if new_parents:
                parent_commit = r[commit.parents[0]]
                if isinstance(parent_commit, Commit):
                    parent_tree = r[parent_commit.tree]
                    if isinstance(parent_tree, Tree):
                        parent_subtree = extract_subtree(
                            r.object_store, parent_tree, prefix
                        )
                        if parent_subtree and parent_subtree.id == commit_subtree.id:
                            # Subtree unchanged, reuse parent's split
                            cache[commit_id] = new_parents[0]
                            return new_parents[0]

            # Create new commit with subtree as root
            new_commit_obj = commit.copy()
            if not isinstance(new_commit_obj, Commit):
                raise TypeError("Expected Commit copy")
            new_commit_obj.tree = commit_subtree.id
            new_commit_obj.parents = new_parents

            r.object_store.add_object(new_commit_obj)
            cache[commit_id] = new_commit_obj.id
            return new_commit_obj.id

        # Process the history
        cache: dict[ObjectID, ObjectID | None] = {}
        split_commit_id = process_commit(head_commit.id, cache)

        if split_commit_id is None:
            raise Error("Failed to split subtree")

        # Create branch if requested
        if branch:
            from dulwich.refs import Ref

            ref_name = Ref(LOCAL_BRANCH_PREFIX + branch)
            r.refs[ref_name] = split_commit_id

        # Rejoin if requested
        if rejoin:
            # Create a merge commit that joins the split back
            rejoin_message = (
                f"Split '{prefix.decode('utf-8', 'replace')}' into branch '{branch.decode('utf-8', 'replace') if branch else 'split'}'"
            ).encode()
            rejoin_message = add_subtree_metadata(
                rejoin_message, prefix, split=split_commit_id
            )

            rejoin_commit_obj = head_commit.copy()
            if not isinstance(rejoin_commit_obj, Commit):
                raise TypeError("Expected Commit copy")
            rejoin_commit_obj.tree = head_tree.id
            rejoin_commit_obj.parents = [head_commit.id, split_commit_id]
            rejoin_commit_obj.author_time = int(time.time())
            rejoin_commit_obj.commit_time = int(time.time())
            rejoin_commit_obj.message = rejoin_message

            r.object_store.add_object(rejoin_commit_obj)
            r[b"HEAD"] = rejoin_commit_obj.id

        return split_commit_id


def subtree_pull(
    repo: RepoPath,
    prefix: bytes | str,
    repository: str,
    ref: bytes | str,
    squash: bool = False,
    message: bytes | None = None,
) -> ObjectID:
    """Pull changes from a remote repository into a subtree.

    Args:
      repo: Repository to pull into
      prefix: Directory path of the subtree
      repository: URL of repository to pull from
      ref: Ref to fetch
      squash: If True, squash history into a single commit
      message: Custom commit message

    Returns:
      Object ID of the merge commit

    Raises:
      Error: If fetch or merge fails
    """
    from . import fetch, open_repo_closing

    if isinstance(prefix, str):
        prefix = prefix.encode("utf-8")
    if isinstance(ref, str):
        ref = ref.encode("utf-8")

    with open_repo_closing(repo) as r:
        # Fetch from remote
        from io import StringIO

        from dulwich.refs import Ref

        fetch_result = fetch(r, repository, outstream=StringIO())

        # Get the fetched commit
        if ref.startswith(b"refs/"):
            ref_key = Ref(ref)
        else:
            ref_key = Ref(LOCAL_BRANCH_PREFIX + ref)

        commit_id = fetch_result.refs.get(ref_key)
        if commit_id is None:
            # Try as a commit SHA - ref itself might be a commit SHA
            merge_commit = parse_commit(r, ref)
            return subtree_merge(
                repo, prefix, merge_commit, squash=squash, message=message
            )

        # Merge the fetched commit
        merge_commit_obj = r[commit_id]
        if not isinstance(merge_commit_obj, Commit):
            raise NotCommitError(commit_id)
        return subtree_merge(
            repo, prefix, merge_commit_obj, squash=squash, message=message
        )


def subtree_push(
    repo: RepoPath,
    prefix: bytes | str,
    repository: str,
    ref: bytes | str,
) -> None:
    """Push a subtree to a remote repository.

    Args:
      repo: Repository to push from
      prefix: Directory path of the subtree
      repository: URL of repository to push to
      ref: Ref to push to

    Raises:
      Error: If split or push fails
    """
    from . import open_repo_closing, push

    if isinstance(prefix, str):
        prefix = prefix.encode("utf-8")
    if isinstance(ref, str):
        ref = ref.encode("utf-8")

    with open_repo_closing(repo) as r:
        # Split the subtree
        split_commit_id = subtree_split(repo, prefix)

        # Push the split commit to the remote
        if not ref.startswith(b"refs/"):
            ref = LOCAL_BRANCH_PREFIX + ref

        # Create a refspec for pushing
        refspec = split_commit_id + b":" + ref
        push(r, repository, refspec)



# file.py -- Safe access to git files
# Copyright (C) 2010 Google, Inc.
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Safe access to git files."""

__all__ = [
    "FileLocked",
    "GitFile",
    "ensure_dir_exists",
]

import os
import sys
import warnings
from collections.abc import Iterable, Iterator
from types import TracebackType
from typing import IO, Any, ClassVar, Literal, overload

from ._typing import Buffer

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


def ensure_dir_exists(
    dirname: str | bytes | os.PathLike[str] | os.PathLike[bytes],
) -> None:
    """Ensure a directory exists, creating if necessary."""
    try:
        os.makedirs(dirname)
    except FileExistsError:
        pass


def _fancy_rename(oldname: str | bytes, newname: str | bytes) -> None:
    """Rename file with temporary backup file to rollback if rename fails."""
    if not os.path.exists(newname):
        os.rename(oldname, newname)
        return

    # Defer the tempfile import since it pulls in a lot of other things.
    import tempfile

    # destination file exists
    (fd, tmpfile) = tempfile.mkstemp(".tmp", prefix=str(oldname), dir=".")
    os.close(fd)
    os.remove(tmpfile)
    os.rename(newname, tmpfile)
    try:
        os.rename(oldname, newname)
    except OSError:
        os.rename(tmpfile, newname)
        raise
    os.remove(tmpfile)


@overload
def GitFile(
    filename: str | bytes | os.PathLike[str] | os.PathLike[bytes],
    mode: Literal["wb"],
    bufsize: int = -1,
    mask: int = 0o644,
    fsync: bool = True,
) -> "_GitFile": ...


@overload
def GitFile(
    filename: str | bytes | os.PathLike[str] | os.PathLike[bytes],
    mode: Literal["rb"] = "rb",
    bufsize: int = -1,
    mask: int = 0o644,
    fsync: bool = True,
) -> IO[bytes]: ...


@overload
def GitFile(
    filename: str | bytes | os.PathLike[str] | os.PathLike[bytes],
    mode: str = "rb",
    bufsize: int = -1,
    mask: int = 0o644,
    fsync: bool = True,
) -> "IO[bytes] | _GitFile": ...


def GitFile(
    filename: str | bytes | os.PathLike[str] | os.PathLike[bytes],
    mode: str = "rb",
    bufsize: int = -1,
    mask: int = 0o644,
    fsync: bool = True,
) -> "IO[bytes] | _GitFile":
    """Create a file object that obeys the git file locking protocol.

    Returns: a builtin file object or a _GitFile object

    Note: See _GitFile for a description of the file locking protocol.

    Only read-only and write-only (binary) modes are supported; r+, w+, and a
    are not.  To read and write from the same file, you can take advantage of
    the fact that opening a file for write does not actually open the file you
    request.

    The default file mask makes any created files user-writable and
    world-readable.

    Args:
      filename: Path to the file
      mode: File mode (only 'rb' and 'wb' are supported)
      bufsize: Buffer size for file operations
      mask: File mask for created files
      fsync: Whether to call fsync() before closing (default: True)

    """
    if "a" in mode:
        raise OSError("append mode not supported for Git files")
    if "+" in mode:
        raise OSError("read/write mode not supported for Git files")
    if "b" not in mode:
        raise OSError("text mode not supported for Git files")
    if "w" in mode:
        return _GitFile(filename, mode, bufsize, mask, fsync)
    else:
        return open(filename, mode, bufsize)


class FileLocked(Exception):
    """File is already locked."""

    def __init__(
        self,
        filename: str | bytes,
        lockfilename: str | bytes,
    ) -> None:
        """Initialize FileLocked.

        Args:
          filename: Name of the file that is locked
          lockfilename: Name of the lock file
        """
        self.filename = filename
        self.lockfilename = lockfilename
        super().__init__(filename, lockfilename)


class _GitFile(IO[bytes]):
    """File that follows the git locking protocol for writes.

    All writes to a file foo will be written into foo.lock in the same
    directory, and the lockfile will be renamed to overwrite the original file
    on close.

    Note: You *must* call close() or abort() on a _GitFile for the lock to be
        released. Typically this will happen in a finally block.
    """

    _file: IO[bytes]
    _filename: str | bytes
    _lockfilename: str | bytes
    _closed: bool

    PROXY_PROPERTIES: ClassVar[set[str]] = {
        "encoding",
        "errors",
        "mode",
        "name",
        "newlines",
        "softspace",
    }
    PROXY_METHODS: ClassVar[set[str]] = {
        "__iter__",
        "__next__",
        "flush",
        "fileno",
        "isatty",
        "read",
        "readable",
        "readline",
        "readlines",
        "seek",
        "seekable",
        "tell",
        "truncate",
        "writable",
        "write",
        "writelines",
    }

    def __init__(
        self,
        filename: str | bytes | os.PathLike[str] | os.PathLike[bytes],
        mode: str,
        bufsize: int,
        mask: int,
        fsync: bool = True,
    ) -> None:
        # Convert PathLike to str/bytes for our internal use
        self._filename: str | bytes = os.fspath(filename)
        self._fsync = fsync
        if isinstance(self._filename, bytes):
            self._lockfilename: str | bytes = self._filename + b".lock"
        else:
            self._lockfilename = self._filename + ".lock"
        try:
            fd = os.open(
                self._lockfilename,
                os.O_RDWR | os.O_CREAT | os.O_EXCL | getattr(os, "O_BINARY", 0),
                mask,
            )
        except FileExistsError as exc:
            raise FileLocked(self._filename, self._lockfilename) from exc
        self._file = os.fdopen(fd, mode, bufsize)
        self._closed = False

    def __iter__(self) -> Iterator[bytes]:
        """Iterate over lines in the file."""
        return iter(self._file)

    def abort(self) -> None:
        """Close and discard the lockfile without overwriting the target.

        If the file is already closed, this is a no-op.
        """
        if self._closed:
            return
        self._file.close()
        try:
            os.remove(self._lockfilename)
            self._closed = True
        except FileNotFoundError:
            # The file may have been removed already, which is ok.
            self._closed = True

    def close(self) -> None:
        """Close this file, saving the lockfile over the original.

        Note: If this method fails, it will attempt to delete the lockfile.
            However, it is not guaranteed to do so (e.g. if a filesystem
            becomes suddenly read-only), which will prevent future writes to
            this file until the lockfile is removed manually.

        Raises:
          OSError: if the original file could not be overwritten. The
            lock file is still closed, so further attempts to write to the same
            file object will raise ValueError.
        """
        if self._closed:
            return
        self._file.flush()
        if self._fsync:
            os.fsync(self._file.fileno())
        self._file.close()
        try:
            if getattr(os, "replace", None) is not None:
                os.replace(self._lockfilename, self._filename)
            else:
                if sys.platform != "win32":
                    os.rename(self._lockfilename, self._filename)
                else:
                    # Windows versions prior to Vista don't support atomic
                    # renames
                    _fancy_rename(self._lockfilename, self._filename)
        finally:
            self.abort()

    def __del__(self) -> None:
        if not getattr(self, "_closed", True):
            warnings.warn(f"unclosed {self!r}", ResourceWarning, stacklevel=2)
            self.abort()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if exc_type is not None:
            self.abort()
        else:
            self.close()

    def __fspath__(self) -> str | bytes:
        """Return the file path for os.fspath() compatibility."""
        return self._filename

    @property
    def closed(self) -> bool:
        """Return whether the file is closed."""
        return self._closed

    def __getattr__(self, name: str) -> Any:  # noqa: ANN401
        """Proxy property calls to the underlying file."""
        if name in self.PROXY_PROPERTIES:
            return getattr(self._file, name)
        raise AttributeError(name)

    # Implement IO[bytes] methods by delegating to the underlying file
    def read(self, size: int = -1) -> bytes:
        return self._file.read(size)

    # TODO: Remove type: ignore when Python 3.10 support is dropped (Oct 2026)
    # Python 3.10 has issues with IO[bytes] overload signatures
    def write(self, data: Buffer, /) -> int:  # type: ignore[override,unused-ignore]
        return self._file.write(data)

    def readline(self, size: int = -1) -> bytes:
        return self._file.readline(size)

    def readlines(self, hint: int = -1) -> list[bytes]:
        return self._file.readlines(hint)

    # TODO: Remove type: ignore when Python 3.10 support is dropped (Oct 2026)
    # Python 3.10 has issues with IO[bytes] overload signatures
    def writelines(self, lines: Iterable[Buffer], /) -> None:  # type: ignore[override,unused-ignore]
        return self._file.writelines(lines)

    def seek(self, offset: int, whence: int = 0) -> int:
        return self._file.seek(offset, whence)

    def tell(self) -> int:
        return self._file.tell()

    def flush(self) -> None:
        return self._file.flush()

    def truncate(self, size: int | None = None) -> int:
        return self._file.truncate(size)

    def fileno(self) -> int:
        return self._file.fileno()

    def isatty(self) -> bool:
        return self._file.isatty()

    def readable(self) -> bool:
        return self._file.readable()

    def writable(self) -> bool:
        return self._file.writable()

    def seekable(self) -> bool:
        return self._file.seekable()

    def __next__(self) -> bytes:
        return next(iter(self._file))
